"""
Base/Blueprint class for cache implementation
"""
from abc import ABC, abstractmethod
from typing import Optional, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ...pyjolt import PyJolt

class BaseCacheBackend(ABC):
    """
    Abstract cache backend blueprint.

    Subclasses should implement:
    - configure_from_app(cls, app) -> BaseCacheBackend
    - connect / disconnect
    - get / set / delete / clear
    """

    @classmethod
    @abstractmethod
    def configure_from_app(cls, app: "PyJolt", configs: dict[str, Any]) -> "BaseCacheBackend":
        """Create a configured backend instance using app config."""

    @abstractmethod
    async def connect(self) -> None:
        """Establish any required connections (no-op for memory)."""

    @abstractmethod
    async def disconnect(self) -> None:
        """Tear down connections (no-op for memory)."""

    @abstractmethod
    async def get(self, key: str) -> Optional[dict]:
        """Return cached payload dict or None."""

    @abstractmethod
    async def set(self, key: str, value: dict, duration: Optional[int] = None) -> None:
        """Store payload dict under key with optional TTL in seconds."""

    @abstractmethod
    async def delete(self, key: str) -> None:
        """Delete a cached entry if present."""

    @abstractmethod
    async def clear(self) -> None:
        """Clear the entire cache namespace."""



"""Admin dashboard extension"""
from __future__ import annotations
import os
from typing import TYPE_CHECKING, Optional, Type, Any, cast, TypedDict, NotRequired
from pydantic import BaseModel, Field
from wtforms_sqlalchemy.orm import model_form
from ..exceptions.runtime_exceptions import CustomException
from .utilities import FormType
from ..base_extension import BaseExtension
from .admin_controller import AdminController
from .database_controller import AdminDatabaseController
from .email_clients_controller import AdminEmailClientsController
from .task_managers_controller import AdminTaskManagersController
from .file_controller import AdminFileController
from ..database.sql.declarative_base import DeclarativeBaseModel
from ..controller import path
from ..request import Request
from ..database.sql import SqlDatabase, AsyncSession
from ..email.email_client import EmailClient
from ..task_manager import TaskManager
from ..utilities import to_kebab_case
from ..caching import Cache

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

class _AdminDashboardConfig(BaseModel):
    """Admin dashboard configuration model."""
    model_config = {"strict": True}

    DASHBOARD_URL: Optional[str] = Field(
        "/admin/dashboard",
        description="URL path for accessing the admin dashboard."
    )
    LOGO_URL: Optional[str] = Field(
        None,
        description="URL for the logo of the admin dashboard displayed on the login page and dashboard menu."
    )
    URL_FOR_FOR_LOGIN: str = Field(description="The url_for string for your login endpoint")
    URL_FOR_FOR_LOGOUT: str = Field(description="The url_for string for your logout endpoint")
    URL_FOR_FOR_PASSWORD_RESET: Optional[str] = Field(None, description="The url_for string for your password reset endpoint.")
    USE_REMEMBER_ME: Optional[bool] = Field(False, description="If a 'Remember me' field should be provided in the lgin form for permanent/long-running cookies/sessions")

class AdminConfig(TypedDict):
    """Admin configurations typed dictionary"""
    DASHBOARD_URL: NotRequired[str]
    LOGO_URL: NotRequired[str]
    URL_FOR_FOR_LOGIN: str
    URL_FOR_FOR_LOGOUT: str
    URL_FOR_FOR_PASSWORD_RESET: NotRequired[str]
    USE_REMEMBER_ME: NotRequired[bool]

class AdminMissingDatabaseExtension(CustomException):
    def __init__(self, db_name: str):
        self.message = ("Failed to load database extension with "
                        f"{db_name=}")

class AdminDashboard(BaseExtension):
    """Admin dashboard extension class."""

    def __init__(self) -> None:
        self._databases_models: dict[str, list[Type[DeclarativeBaseModel]]] = {}
        self._configs: dict[str, Any] = {}
        self._configs_name: str = "ADMIN_DASHBOARD"
        self._root_path = os.path.dirname(__file__)
        self._databases: dict[str, SqlDatabase] = {}
        self._email_clients: Optional[dict[str, EmailClient]]
        self._task_managers: Optional[dict[str, TaskManager]]

    def init_app(self, app: "PyJolt") -> None:
        self._app = app
        self._configs = app.get_conf(self._configs_name, {})
        self._configs = self.validate_configs(self._configs, _AdminDashboardConfig)
        #pylint: disable-next=W0212
        self._databases_models = self.get_registered_models()
        self._databases = self._get_all_databases()
        self._email_clients = self.get_email_clients()
        self._task_managers = self.get_task_managers()
        self._app.add_template_path(self._root_path)

        for ctrl in [AdminController, AdminDatabaseController,
                     AdminEmailClientsController, AdminTaskManagersController,
                     AdminFileController]:
            ctrl = path(url_path=self._configs["DASHBOARD_URL"],
                                                 open_api_spec=False)(ctrl)
            setattr(ctrl, "_dashboard", self)
            self._app.register_controller(ctrl)

    def get_model(self, db_name: str, model_name: str) -> Type[DeclarativeBaseModel] | None:
        """Get a model class by database name and model name."""
        models = self._databases_models.get(db_name, [])
        for model in models:
            if model.__name__ == model_name:
                return model
        return None
    
    def get_model_form(self, model: Type[DeclarativeBaseModel],
                       form_type: str = FormType.UPDATE, 
                       exclude_pk: bool = False, exclude_fk: bool = True,
                       only: Any | None = None, exclude: Any | None = None,
                       field_args: Any | None = None, converter: Any | None = None,
                       type_name: Any | None = None) -> dict[str, Any]:#Type:
        """
            Get a WTForms-SQLAlchemy form for a given model.

            Args:
                only (Iterable[str], optional):  
                    Property names that should be included in the form.  
                    Only these properties will have fields.

                exclude (Iterable[str], optional):  
                    Property names that should be excluded from the form.  
                    All other properties will have fields.

                field_args (dict[str, dict], optional):  
                    A mapping of field names to keyword arguments used to construct
                    each field object.

                converter (Type[ModelConverter], optional):  
                    A converter class used to generate fields based on the model
                    properties. If not provided, ``ModelConverter`` is used.

                exclude_pk (bool, optional):  
                    Whether to force exclusion of primary key fields. Defaults to ``False``.

                exclude_fk (bool, optional):  
                    Whether to force exclusion of foreign key fields. Defaults to ``False``.

                type_name (str, optional):  
                    Custom name for the generated form class.

            Returns:
                Type[Form]: A dynamically generated WTForms form class.
        """
        if hasattr(model, f"__{form_type}_form__"):
            return getattr(model, f"__{form_type}_form__")
        form_class = model_form(model, exclude_pk=exclude_pk, only=only, exclude=exclude,
                                field_args=field_args, converter=converter,
                                exclude_fk=exclude_fk, type_name=type_name)
        setattr(model, f"__{form_type}_form__", form_class)
        return form_class

    def get_database(self, db_name: str) -> "SqlDatabase":
        for _, ext in self._databases.items():
            if ext.db_name == db_name:
                return ext
        raise AdminMissingDatabaseExtension(db_name)
    
    def _get_all_databases(self) -> dict[str, SqlDatabase]:
        """Gets all database extensions from app extensions"""
        databases: dict[str, SqlDatabase] = {}
        for _, ext in self.app.extensions.items():
            if isinstance(ext, SqlDatabase) and self._databases_models.get(ext.configs_name, None) is not None:
                databases[ext.db_name] = ext
        return databases
    
    def get_registered_models(self) -> dict[str, list[Type[DeclarativeBaseModel]]]:
        """Gets all registered models for admin dashboard"""
        #get_registered_models()#self._app._db_models
        databases_and_models: dict[str, list[Type[DeclarativeBaseModel]]] = {}
        #pylint: disable-next=W0212
        for db_name, models in cast("PyJolt", self._app)._db_models.items():
            registered_models: list[Type[DeclarativeBaseModel]] = []
            for m in models:
                if hasattr(m, "__use_in_dashboard__") and getattr(m, "__use_in_dashboard__", False) is True:
                    registered_models.append(m)
            if len(registered_models) > 0:
                #pylint: disable-next=W0212
                databases_and_models[self.app._db_name_configs_map[db_name]] = registered_models
        return databases_and_models
    
    def get_session(self, database: "SqlDatabase") -> "AsyncSession":
        return database.create_session()
    
    async def number_of_tables(self) -> int:
        """Number of all tables in all databases"""
        num: int = 0
        for _, db in self._databases.items():
            num = num + await db.count_tables()
        return num
    
    async def number_of_rows(self) -> int:
        """Number of all rows in all databases"""
        num: int = 0
        for _, db in self._databases.items():
            _, rows = await db.count_rows_exact()
            num = num + rows
        return num
    
    async def databases_overviews(self, with_extras=False) -> dict[str, Any]:
        """Collects database overviews for dashboard"""
        overviews: dict[str, Any] = {
            "db_count": 0,
            "schemas_count": 0,
            "tables_count": 0,
            "views_count": 0,
            "columns_count": 0,
            "rows_count": 0
        }
        overviews["db_count"] = self.number_of_dbs
        for _, db in self._databases.items():
            overview = await db.collect_db_overview(with_extras=with_extras)
            _, rows_count = await db.count_rows_exact()
            overview["rows_count"] = rows_count
            overviews["schemas_count"]+=overview["schemas_count"]
            overviews["tables_count"]+=overview["tables_count"]
            overviews["views_count"]+=overview["views_count"]
            overviews["columns_count"]+=overview["columns_count"]
            overviews["rows_count"]+=overview["rows_count"]
        return overviews
    
    async def database_overview(self, db_name: str, with_extras: bool = False) -> dict[str, Any]:
        """Returns overview for selected db"""
        db: SqlDatabase = cast(SqlDatabase, self._databases.get(db_name))
        if db is None:
            #pylint: disable-next=W0719
            raise Exception(f"Unknown database: {db_name}")
        overview: dict[str, Any] = await db.collect_db_overview(with_extras=with_extras)
        _, rows_count = await db.count_rows_exact()
        overview["rows_count"] = rows_count
        return overview
    
    def get_email_clients(self) -> Optional[dict[str, EmailClient]]:
        """Finds all registered email client extensions"""
        clients: dict[str, EmailClient] = {}
        for _, ext in self.app.extensions.items():
            if isinstance(ext, EmailClient):
                clients[to_kebab_case(ext.configs_name)] = ext
        if len(clients.keys())==0:
            return None
        return clients
    
    def get_task_managers(self) -> Optional[dict[str, TaskManager]]:
        """Finds all registered task managers"""
        task_managers: dict[str, TaskManager] = {}
        for _, ext in self.app.extensions.items():
            if isinstance(ext, TaskManager):
                task_managers[ext.configs_name] = ext
        if len(task_managers.keys()) == 0:
            return None
        return task_managers
    
    def get_cache_interfaces(self) -> Optional[dict[str, Cache]]:
        """Cache extensions"""
        caches: dict[str, Cache] = {}
        for _, ext in self.app.extensions.items():
            if isinstance(ext, Cache):
                caches[ext.configs_name] = ext
        if len(caches.keys()) == 0:
            return None
        return caches

    async def email_recipient_query(self, req: Request, query: str,
                                    client: EmailClient) -> list[tuple[str, str]]:
        """
        Email recipients query method. Should return a list of tuples
        containing name-value pairs. Example: [("John Doe", "john.doe@email.com)]
        If you wish you can also simply return an empty list

        Example:
        return [
            ("John Doe", "john.doe@email.com"),
            ("Jane Doe", "jane.doe@email.com"),
            ("John Smith", "john.smith@email.com"),
            ("Alexander The Great", "alexander.great@email.com"),
            ("Julius Cesar", "julius.cesar@email.com"),
            ("Marcus Antonius", "marcus.antonius@email.com")
        ]
        """
        raise NotImplementedError("Please override the 'email_recipient_query' "
                                   "method in your AdminDashboard implementation "
                                   "for this functionality to work.")

    @property
    def root_path(self) -> str:
        return self._root_path

    @property
    def email_clients(self) -> Optional[dict[str, EmailClient]]:
        """Dictionary of email clients"""
        return self._email_clients

    @property
    def task_managers(self) -> Optional[dict[str, TaskManager]]:
        """Dictionary of task managers"""
        return self._task_managers

    @property
    def number_of_dbs(self) -> int:
        """Number of databases"""
        return len(self._databases)

    @property
    def all_dbs(self) -> list[SqlDatabase]:
        """List of all databases"""
        return [db for db in self._databases.values()]

    async def has_enter_permission(self, req: Request) -> bool:
        """If a user can enter the dashboard"""
        raise NotImplementedError("Please implement 'has_enter_permission' method "
                                  "before using the admin dashboard")

    async def has_view_permission(self, req: Request, model: Type[DeclarativeBaseModel]) -> bool:
        """If the logged in user has permission to view model data"""
        raise NotImplementedError("Please implement 'has_view_permission' method "
                                  "for viewing database data/models before "
                                  "using the admin dashboard.")

    async def has_update_permission(self, req: Request, model: Type[DeclarativeBaseModel]) -> bool:
        """If the logged in user has permission to edit model data"""
        raise NotImplementedError("Please implement 'has_update_permission' method "
                                  "for updating database data/models before "
                                  "using the admin dashboard.")

    async def has_create_permission(self, req: Request, model: Type[DeclarativeBaseModel]) -> bool:
        """If the logged in user has permission to create model data"""
        raise NotImplementedError("Please implement 'has_create_permission' method "
                                  "for creating database model records before "
                                  "using the admin dashboard.")

    async def has_delete_permission(self, req: Request, model: Type[DeclarativeBaseModel]) -> bool:
        """If the logged in user has permission to delete model data"""
        raise NotImplementedError("Please implement 'has_delete_permission' method "
                                  "for deleting database before "
                                  "using the admin dashboard.")

    async def has_email_permission(self, req: Request, client: EmailClient) -> bool:
        """If the logged in user has permission to use email clients"""
        raise NotImplementedError("Please implement the 'has_email_permission' "
                                  "method before using the email client extension "
                                  "in the admin dashboard.")

    async def has_task_manager_permission(self, req: Request, manager: TaskManager) -> bool:
        """If the logged in user has permission to use task managers"""
        raise NotImplementedError("Please implement the 'has_task_manager_permission'"
                                  " method before using the task manager client extensions"
                                  " in the admin dashboard")
    
    async def has_cache_permission(self, req: Request, cache: Cache) -> bool:
        """If the logged in user has permission to reset the cache"""
        raise NotImplementedError("Please implement the 'has_cache_permission'"
                                  " method before using the cache interface extensions"
                                  " in the admin dashboard")
    
    async def has_files_permission(self, req: Request) -> bool:
        """If user can edit/upload files to static assets"""
        raise NotImplementedError("Please implement 'has_files_permission' method "
                                  "before using the admin dashboard")

    async def get_all_permissions(self, req: Request) -> dict[str, Any]:
        """Returns a map of """
        permissions: dict[str, Any] = {}
        for db in self.all_dbs:
            permissions[db.configs_name] = {}
            for model in self.get_registered_models()[db.configs_name]:
                permissions[db.configs_name][model.__name__] = {
                    "CREATE": await self.has_create_permission(req, model),
                    "UPDATE": await self.has_update_permission(req, model),
                    "DELETE": await self.has_delete_permission(req, model),
                    "VIEW": await self.has_view_permission(req, model)
                }
        email_clients = self.get_email_clients()
        if email_clients is not None:
            for name, client in email_clients.items():
                permissions[name] = await self.has_email_permission(req, client)
        
        task_managers = self.get_task_managers()
        if task_managers is not None:
            for name, manager in task_managers.items():
                permissions[name] = await self.has_task_manager_permission(req, manager)
        
        cache_interfaces = self.get_cache_interfaces()
        if cache_interfaces is not None:
            for name, interface in cache_interfaces.items():
                permissions[name] = await self.has_cache_permission(req, interface)

        permissions["files"] = await self.has_files_permission(req)

        return permissions
                



"""
App configurations
"""
import os
from pyjolt import BaseConfig

class Config(BaseConfig):
    """Config class"""
    APP_NAME: str = "Test app"
    VERSION: str = "1.0"
    SECRET_KEY: str = "46373hdnsfshf73462twvdngnghjdgsfd"
    BASE_PATH: str = os.path.dirname(__file__)

    DATABASE_URI: str = "sqlite+aiosqlite:///./test.db"
    ALEMBIC_DATABASE_URI_SYNC: str = "sqlite:///./test.db"

    CONTROLLERS: list[str] = [
        'app.api.example_api:ExampleApi'
    ]

    CLI_CONTROLLERS: list[str] = [
        'app.cli.cli_controller:UtilityCLIController'
    ]

    EXTENSIONS: list[str] = [
        'app.extensions:db',
        'app.extensions:migrate',
        'app.authentication:auth'
    ]

    MODELS: list[str] = [
        'app.api.models:Example'
    ]

    EXCEPTION_HANDLERS: list[str] = [
        'app.api.exceptions.exception_handler:Handler'
    ]



"""
Main admin dashboard controller module.
Handles login and static files
"""
# pylint: disable=W0719,W0212
from __future__ import annotations

import mimetypes
import os
from typing import TYPE_CHECKING, Any, Optional

from werkzeug.security import safe_join

from pyjolt.http_statuses import HttpStatus

from ..controller import get
from ..auth.authentication import login_required
from ..exceptions.http_exceptions import StaticAssetNotFound
from ..request import Request
from ..response import Response
from ..utilities import get_file, get_range_file
from .common_controller import CommonAdminController
from ..caching import Cache

if TYPE_CHECKING:
    from .admin_dashboard import AdminDashboard

class AdminController(CommonAdminController):
    """Admin dashboard controller."""

    _dashboard: "AdminDashboard"

    @get("/")
    @login_required
    async def index(self, req: Request) -> Response:
        """Index page of dashboard"""
        if not (await self.can_enter(req)):
            return await self.cant_enter_response(req)
        level: str = req.query_params.get("logs", "all")
        logs: list[dict[str, Any]] = self.app.log_buffer.get_all() if level == "all" else self.app.log_buffer.get_severe()
        logs.reverse()
        caches: Optional[dict[str, Cache]] = self.dashboard.get_cache_interfaces()
        cache_permissions: dict[str, bool] = {}
        if caches is not None:
            for key, cache in caches.items():
                cache_permissions[key] = await self.dashboard.has_cache_permission(req, cache)
        return await req.res.html(
            "/__admin_templates/dashboard.html",
            {"configs": self.dashboard.configs,
             "all_logs": logs,
             "caches": caches,
             "cache_permissions": cache_permissions,
             **self.get_common_variables()}
        )

    @get("/cache-reset/<string:cache_name>")
    @login_required
    async def reset_cache(self, req: Request, cache_name: str) -> Response:
        """
        Resets cache with provided cache_name
        """
        if not (await self.can_enter(req)):
            return await self.cant_enter_response(req)

        caches: Optional[dict[str, Cache]] = self.dashboard.get_cache_interfaces()
        if caches is None:
            return req.res.json({
                "message": "No cache extensions found or installed",
                "status": "danger"
            }).status(HttpStatus.BAD_REQUEST)

        cache: Optional[Cache] = caches.get(cache_name)
        if cache is None:
            return req.res.json({
                "message": f"No cache extensions matching name {cache_name} found",
                "status": "warning"
            }).status(HttpStatus.NOT_FOUND)

        if not (await self.dashboard.has_cache_permission(req, cache)):
            return req.res.json({
                "message": "User doesn't have permission to reset the cache",
                "status": "danger"
            }).status(HttpStatus.UNAUTHORIZED)

        await cache.clear()

        return req.res.json({
            "message": "Cache reset successful",
            "status": "success"
        }).status(HttpStatus.OK)

    @get("/login")
    async def login(self, req: Request) -> Response:
        """Login route for dashboard"""
        return await req.res.html(
            "/__admin_templates/login.html",
            {"configs": self.dashboard.configs}
        )

    # @get("/stream")
    # async def stream_example(self, req: "Request") -> Response[bytes]:
    #     async def gen():
    #         for i in range(10):
    #             yield f"data: {i}\n\n"
    #             await asyncio.sleep(1)

    #     return req.res.stream_text(gen())

    #STATIC files for dashboard
    @get("/static/<path:filename>")
    async def static(self, req: Request, filename: str) -> Response:
        """Serves static assets for the dashboard"""
        file_path = None
        candidate = safe_join(self.dashboard.root_path, "static", filename)
        if candidate and os.path.exists(candidate):
            file_path = candidate
        if not file_path:
            raise StaticAssetNotFound()

        # checks/guesses mimetype
        guessed, _ = mimetypes.guess_type(file_path)
        content_type = guessed or "application/octet-stream"

        # Checks range header and returns range if header is present
        range_header = req.headers.get("range")
        if not range_header:
            status, headers, body = await get_file(file_path, content_type=content_type)
            headers["Accept-Ranges"] = "bytes"
            return req.res.send_file(body, headers).status(status)

        return await get_range_file(req.res, file_path, range_header, content_type)



"""Controller for state machine extension"""
from enum import Enum, IntEnum, StrEnum
from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel, Field

from pyjolt.http_statuses import HttpStatus

from ..controller import Controller, post
from ..request import Request
from ..response import Response

if TYPE_CHECKING:
    from .state_machine import StateMachine

class TransitionRequestData(BaseModel):
    """Request model for state transition"""
    step: Enum|StrEnum|IntEnum = Field(description="Step to perform")
    data: Optional[dict[str, Any]] = Field(None,
                        description="Additional data for the transition")

class StateMachineController(Controller):
    """Controller for state machine operations"""

    state_machine: "StateMachine"

    @post("/transition")
    async def transition_state(self, req: Request) -> Response:
        """
        Transition to the next state based on the current state.
        """
        data = await req.json()
        if data is None:
            return req.res.json({
                "message": "Transition data is missing",
                "status": "error"
            }).status(HttpStatus.BAD_REQUEST)
        step = data.pop("step", None)
        if step is None:
            return req.res.json({
                "message": "Missing step information",
                "status": "error"
            }).status(HttpStatus.BAD_REQUEST)

        #pylint: disable-next=W0212
        if step in self.state_machine.steps._member_names_:
            step = self.state_machine.steps[step]
        else:
            return req.res.json({
                "message": f"Step {step} does not exist.",
                "status": "error"
            }).status(HttpStatus.BAD_REQUEST)

        transition_request = self.state_machine.transition_request_data(step=step,
                                                                **data).model_dump()
        current_state, transition_context = await self.state_machine.context_loader(
                                                    req,
                                                    transition_request)
        transition_request["current_state"] = current_state
        method = self.state_machine.step_methods_map.get(step, None)
        if method is None:
            return req.res.json({
                "message": f"Method for step {step.name} not found in mapping.",
                "status": "error"
            }).status(HttpStatus.NOT_FOUND)
        next_state = self.state_machine.get_next_state(current_state, step)
        if next_state is None:
            return req.res.json({
                "message": f"Step {step.name} not allowed from current state {current_state.name}",
                "status": "error"
            }).status(HttpStatus.BAD_REQUEST)
        return await method(req, transition_request, transition_context, next_state)



"""
Task manager class
"""
from typing import (Callable, Tuple, Optional,
                    cast, TYPE_CHECKING, Any,
                    TypedDict, NotRequired)
from functools import wraps

from apscheduler.job import Job
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.jobstores.base import JobLookupError
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.executors.asyncio import AsyncIOExecutor
from pydantic import BaseModel, Field

from ..utilities import run_sync_or_async, run_in_background
from ..base_extension import BaseExtension

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

class _TaskManagerConfigs(BaseModel):
    """Configuration model for TaskManager extension."""
    NICE_NAME: Optional[str] = Field("Task manager", description="Human readable name for the task manager for the admin dashboard")
    SCHEDULER: Optional[Callable] = Field(
        default=AsyncIOScheduler,
        description="Scheduler class to use, must be subclass of BaseScheduler"
    )
    JOB_STORES: Optional[dict] = Field(
        default={"default": MemoryJobStore()},
        description="Job stores configuration dictionary"
    )
    EXECUTORS: Optional[dict] = Field(
        default={"default": AsyncIOExecutor()},
        description="Executors configuration dictionary"
    )
    JOB_DEFAULTS: Optional[dict] = Field(
        default={
            'coalesce': False,
            'max_instances': 3
        },
        description="Default job settings dictionary"
    )
    DAEMON: Optional[bool] = Field(
        default=True,
        description="Whether the scheduler should run as a daemon"
    )

class TaskManagerConfig(TypedDict):
    """TypedDict for TaskManager configuration."""
    NICE_NAME: NotRequired[str]
    SCHEDULER: NotRequired[Callable]
    JOB_STORES: NotRequired[dict]
    EXECUTORS: NotRequired[dict]
    JOB_DEFAULTS: NotRequired[dict]
    DAEMON: NotRequired[bool]

class TaskManager(BaseExtension):
    """
    Task manager class for scheduling and managing backgroudn tasks.
    """

    def __init__(self, configs_name: str = "TASK_MANAGER") -> None:
        self._configs_name: str = cast(str, configs_name)
        self._configs: dict[str, Any] = {}
        self._app: "PyJolt"
        self._job_stores: dict
        self._executors: dict
        self._job_defaults: dict
        self._daemon: bool
        self._scheduler: AsyncIOScheduler
        self._initial_jobs_methods_list: list[Tuple] = []
        self._active_jobs: dict[str, Job] = {}

    def init_app(self, app: "PyJolt"):
        """
        Initlizer for TaskManager with PyJolt app
        """
        self._app = app
        self._configs = app.get_conf(self._configs_name, {})

        self._configs = self.validate_configs(self._configs, _TaskManagerConfigs)
        self._job_stores = self._configs["JOB_STORES"]
        self._executors = self._configs["EXECUTORS"]
        self._job_defaults = self._configs["JOB_DEFAULTS"]
        self._daemon = self._configs["DAEMON"]
        self._scheduler = self._configs["SCHEDULER"]

        self._scheduler = self._scheduler(jobstores=self._job_stores,
                                            executors=self._executors,
                                            job_defaults=self._job_defaults,
                                            daemon=self._daemon
                                            ) #type: ignore
        self._app.add_extension(self)
        self._get_defined_jobs()
        self._app.add_on_startup_method(self._start_scheduler)
        self._app.add_on_shutdown_method(self._stop_scheduler)

    def pause_scheduler(self):
        """
        Pauses scheduler execution
        """
        self.scheduler.pause()

    def resume_scheduler(self):
        """
        Resumes paused scheduler execution
        """
        self.scheduler.resume()

    async def _start_scheduler(self):
        """
        On startup hook for starting the scheduler
        """
        self.scheduler.start()
        self._start_initial_jobs()

    async def _stop_scheduler(self):
        """
        On shutdown hook for shuting the scheduler down
        """
        self.scheduler.shutdown()
    
    def _get_defined_jobs(self):
        for name in dir(self):
            method = getattr(self, name)
            if not callable(method):
                continue
            scheduler_method = getattr(method, "_scheduler_job", None)
            if scheduler_method:
                self._initial_jobs_methods_list.append((method,
                                                        scheduler_method["args"],
                                                        scheduler_method["kwargs"]))

    def _start_initial_jobs(self) -> None:
        """
        Starts all initial jobs (decorated functions)
        """
        if self._initial_jobs_methods_list is None or len(self._initial_jobs_methods_list)==0:
            return
        for func, args, kwargs in self._initial_jobs_methods_list:
            job: Job = self.scheduler.add_job(func, *args, **kwargs)
            self._active_jobs[job.id] = job
        self._initial_jobs_methods_list = []

    def run_background_task(self, func: Callable, *args, **kwargs):
        """
        Runs a method in the background (fire and forget).
        Used for running functions whose execution doesn't have to be awaited/returned to the user.
        Example: a route handler can return a response immediately, and the
        task is executed in a seperate thread (sending an email for example).

        Uses the pyjolt.utilities.run_in_background method
        """
        run_in_background(func, *args, **kwargs)

    def add_job(self, func: Callable, *args, **kwargs) -> Job:
        """
        Adds job
        """
        job: Job = self.scheduler.add_job(func, *args, **kwargs)
        self._active_jobs[job.id] = job
        return job

    def remove_job(self, job: str|Job, job_store: Optional[str] = None):
        """
        Removes a job.
        :param job: job id (str) or the Job instance returned by the scheduler.add_job method
        """
        if isinstance(job, Job):
            job = job.id
        return self._remove_job(cast(str,job), job_store)

    def pause_job(self, job: str|Job):
        """
        Pauses the job
        """
        if isinstance(job, Job):
            return job.pause()
        active_job: Optional[Job] = self._active_jobs.get(job, None)
        if job is None:
            raise JobLookupError(job)
        return cast(Job, active_job).pause()

    def resume_job(self, job: str|Job):
        """
        Resumes job
        :param paused_job: id or Job instance
        """
        if isinstance(job, Job):
            return job.resume()
        paused_job: Optional[Job] = self._active_jobs.get(job, None)
        if paused_job is None:
            raise JobLookupError(paused_job)
        return paused_job.resume()
    
    def get_job(self, job_id: str) -> Job|None:
        return self._active_jobs.get(job_id, None)

    def _remove_job(self, job_id: str, job_store = None):
        """
        Removes job from job list
        """
        self.scheduler.remove_job(job_id, job_store)
        del self._active_jobs[job_id]

    @property
    def jobs(self) -> dict[str, Job]:
        """
        Returns dictionary of running jobs
        """
        return self._active_jobs

    @property
    def scheduler(self) -> AsyncIOScheduler:
        """
        Returns the background scheduler instance
        """
        return self._scheduler
    
    @property
    def nice_name(self) -> str:
        """Nice name of the instance"""
        return self._configs["NICE_NAME"]
    
    @property
    def app(self) -> "PyJolt":
        """
        Application instance
        """
        return cast("PyJolt", self._app)


def schedule_job(*args, **kwargs):
    """
    ```
    A decorator to add a function as a scheduled job in the given APScheduler instance.
    The decorated function is added to a list of tuples (func, args, kwargs) and the job
    is started when the scheduler instance is started (on_startup event of PyJolt)
    IMPORTANT: The decorator should be the top-most decorator of the function to make sure
    any other decorator is applied before the job is added to the job list
    :param args: Positional arguments to pass to scheduler.add_job().
                Typically, the first of these args is the trigger (e.g. 'interval', 'cron', etc.).
    :param kwargs: Keyword arguments to pass to scheduler.add_job().
    Example:
    Runs a job with id 'my_job_id' every 5 minutes

    @schedule_job('interval', minutes=5, id='my_job_id')
    async def my_job(self):
        #some task
    ```
    """
    def decorator(func: Callable):
        @wraps(func)
        async def wrapper(self, *f_args, **f_kwargs):
            return await run_sync_or_async(func, self, *f_args, **f_kwargs)
        setattr(wrapper, "_scheduler_job", {"args": args,
                                            "kwargs": kwargs})
        return wrapper
    return decorator



"""
Admin dashboard database controller
Handles everything database related.
"""
# pylint: disable=W0719,W0212
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Type, cast

from pydantic import BaseModel, Field, ValidationError
from sqlalchemy.inspection import inspect

from ..auth.authentication import login_required
from ..controller import delete, get, post, put
from ..database.sql import AsyncQuery, AsyncSession, SqlDatabase
from ..database.sql.declarative_base import DeclarativeBaseModel
from ..exceptions.http_exceptions import BaseHttpException
from ..http_statuses import HttpStatus
from ..request import Request
from ..response import Response
from .utilities import FormType, PermissionType, extract_table_columns
from .common_controller import CommonAdminController

if TYPE_CHECKING:
    from .input_fields import FormField

class UnknownModelError(BaseHttpException):
    """Unknown model for admin dashboard"""

    def __init__(self, db_name: str, model_name: str):
        """Init method for exception"""
        super().__init__(f"Model {model_name} in database {db_name} does not exist",
                         HttpStatus.NOT_FOUND, "error", {"db_name": db_name,
                                                         "model_name": model_name})

class AdminPermissionError(BaseHttpException):
    """Unknown model for admin dashboard"""

    def __init__(self, user: Any):
        """Init method for exception"""
        super().__init__("User doesn't have permission for this action",
                         HttpStatus.NOT_FOUND, "error", user)

class PaginationModel(BaseModel):
    """Simple pagination model"""
    page: int = Field(1, description="Requested page")
    per_page: int = Field(12, description="Requested number of results per page")

class AdminDatabaseController(CommonAdminController):
    """Admin dashboard controller."""

    @get("/databases")
    @login_required
    async def databases(self, req: Request) -> Response:
        """Get admin dashboard data."""
        if not await self.can_enter(req):
            return await self.cant_enter_response(req)
        overviews: dict[str, Any] = await self.dashboard.databases_overviews()
        return await req.res.html("/__admin_templates/databases.html", {
            "num_of_db": overviews["db_count"], "schemas_count": overviews["schemas_count"],
            "tables_count": overviews["tables_count"],"views_count": overviews["views_count"],
            "rows_count": overviews["rows_count"], "columns_count": overviews["columns_count"],
            **self.get_common_variables()
        })
    
    @get("/database/<string:db_name>")
    @login_required
    async def database(self, req: Request, db_name: str) -> Response:
        """Get admin dashboard data."""
        if not await self.can_enter(req):
            return await self.cant_enter_response(req)
        overview: dict[str, Any] = await self.dashboard.database_overview(db_name, with_extras=True)
        db: SqlDatabase = self.dashboard.get_database(db_name)
        return await req.res.html("/__admin_templates/database.html", {
            "schemas_count": overview["schemas_count"],
            "tables_count": overview["tables_count"],"views_count": overview["views_count"],
            "rows_count": overview["rows_count"], "columns_count": overview["columns_count"],
            "size_bytes": overview.get("extras", {}).get("db_size_bytes", None),
            "db": self.dashboard.get_database(db_name),
            "models_list": db.models_list, "db_name": db_name,
            **self.get_common_variables()
        })

    @get("/data/database/<string:db_name>/model/<string:model_name>")
    @login_required
    async def model_table(self, req: Request, db_name: str,
                                    model_name: str) -> Response:
        """Model table with records."""
        if not await self.can_enter(req):
            return await self.cant_enter_response(req)
        model: Type[DeclarativeBaseModel] = await self.check_permission(
            PermissionType.CAN_VIEW, req, db_name, model_name)
        print("Here")
        pagination = PaginationModel.model_validate(req.query_params)
        database: SqlDatabase = self.dashboard.get_database(db_name)
        session: AsyncSession = self.dashboard.get_session(database)
        query: AsyncQuery = model.query(session)
        ordering: Optional[list[str]] = model.order_table_by()
        if ordering is not None:
            query = query.order_by_strings(
                *ordering
            )
        all_data: dict[str, Any] = {}
        if pagination.per_page > 0:
            all_data= await query.paginate(page=pagination.page, per_page=pagination.per_page)
        else:
            all_data["items"] = await query.all()
            all_data["pages"] = 0
        await session.close()
        columns = extract_table_columns(model, exclude=getattr(model.Meta,
                                                "exclude_from_table", None))
        relationships_tuples = inspect(model).relationships.items()#type: ignore[union-attr]
        relationships: list[str] = []
        if relationships_tuples is not None:
            relationships=[rel[0] for rel in relationships_tuples]
        form: Type[Any]|dict[str, FormField|Any] = self.dashboard.get_model_form(model,
                                                   form_type=FormType.UPDATE,
                                                   exclude_pk = True,
                                                   exclude=relationships
                                                   )
        form = self.generate_form(model, [field for field in cast(Type[Any], form)()])
        create_permission: bool = await self.dashboard.has_create_permission(req, model)
        delete_permission: bool = await self.dashboard.has_delete_permission(req, model)
        update_permission: bool = await self.dashboard.has_update_permission(req, model)
        return await req.res.html(
            "/__admin_templates/model_table.html",
            {"model_name": model_name, "all_data": all_data, "pk_names": model.primary_key_names(),
            "columns": columns, "title": f"{model_name} Table",
            "create_path": self.create_attr_val_path_for_model,
            "db_nice_name": database.nice_name, "db_name": db_name, "db": database,
            "model": model,
            "model_form": form,
            "can_create": create_permission, "can_delete": delete_permission,
            "can_update": update_permission,
            **self.get_common_variables()}
        )

    @get("/data/database/<string:db_name>/model/<string:model_name>/<path:attr_val>")
    @login_required
    async def get_model_record(self, req: Request, db_name: str,
                                    model_name: str, attr_val: str) -> Response:
        """Get a specific model record."""
        if not await self.can_enter(req):
            return await self.cant_enter_response(req)
        model = await self.check_permission(PermissionType.CAN_VIEW, req, db_name, model_name)
        db: SqlDatabase = self.dashboard.get_database(db_name)
        model_pk_val: dict[str, str] = self.parse_key_value_path(attr_val)
        filters = self.get_pk_filters_from_path(model, model_pk_val)
        async with db.create_session() as session:
            async with session.begin():
                record = await model.query(session).filter(*filters).first()
                if record is None:
                    raise UnknownModelError(db_name, model_name)
        data = {}
        for column in model.__table__.columns:
            value = getattr(record, column.key)
            if isinstance(value, datetime):
                value = value.isoformat()
            data[column.key] = value
        return req.res.json({
            "message": "Record data fetched successfully.",
            "status": "success",
            "data": data
        }).status(HttpStatus.OK)

    #API calls for the admin dashboard
    #DELETE, PUT, CREATE operations
    @delete("/data/database/<string:db_name>/model/<string:model_name>/<path:attr_val>")
    @login_required
    async def delete_model_record(self, req: Request, db_name: str,
                                    model_name: str, attr_val: str) -> Response:
        """Delete a specific model record."""
        if not await self.can_enter(req):
            return await self.cant_enter_response(req)
        model = await self.check_permission(PermissionType.CAN_DELETE, req, db_name, model_name)
        db: SqlDatabase = self.dashboard.get_database(db_name)
        model_pk_val: dict[str, str] = self.parse_key_value_path(attr_val)
        filters = self.get_pk_filters_from_path(model, model_pk_val)
        async with db.create_session() as session:
            async with session.begin():
                record = await model.query(session).filter(*filters).first()
                if record is not None:
                    await record.admin_delete(req, session)
                else:
                    return req.res.json({
                        "message": f"Record in {model_name} not found.",
                        "status": "error"
                    }).status(HttpStatus.NOT_FOUND)
        self.app.logger.info(f"Admin dashboard - deleted {model_name=} with {attr_val}")
        return req.res.no_content()

    @put("/data/database/<string:db_name>/model/<string:model_name>/<path:attr_val>")
    @login_required
    async def put_model_record(self, req: Request, db_name: str,
                                    model_name: str, attr_val: str) -> Response:
        """Patch a specific model record."""
        if not await self.can_enter(req):
            return await self.cant_enter_response(req)
        model: Type[DeclarativeBaseModel] = await self.check_permission(PermissionType.CAN_UPDATE, req, db_name, model_name)
        data: dict[str, Any] = cast(dict[str, Any], await req.json())
        validation_schema = model.update_validation_schema()
        if validation_schema is not None:
            try:
                data = validation_schema.model_validate(data).model_dump()
            except ValidationError as exc:
                details = {}
                if hasattr(exc, "errors"):
                    for error in exc.errors():
                        details[error["loc"][0]] = error["msg"]
                return req.response.json({
                    "message": "Validation failed.",
                    "details": details
                }).status(HttpStatus.UNPROCESSABLE_ENTITY)
        db: SqlDatabase = self.dashboard.get_database(db_name)
        model_pk_val: dict[str, str] = self.parse_key_value_path(attr_val)
        filters = self.get_pk_filters_from_path(model, model_pk_val)
        async with db.create_session() as session:
            async with session.begin():
                record: DeclarativeBaseModel = await model.query(session).filter(*filters).first()
                if record is not None:
                    await record.admin_update(req, data, session)
                else:
                    return req.res.json({
                        "message": f"Record in {model_name} not found.",
                        "status": "error"
                    }).status(HttpStatus.NOT_FOUND)
        return req.res.json({
            "message": f"Record in {model_name} updated successfully.",
            "status": "success"
        }).status(HttpStatus.OK)

    @post("/data/database/<string:db_name>/model/<string:model_name>")
    @login_required
    async def create_model_record(self, req: Request, db_name: str, 
                                    model_name: str) -> Response:
        """Create a new model record."""
        if not await self.can_enter(req):
            return await self.cant_enter_response(req)
        model = await self.check_permission(PermissionType.CAN_CREATE, req, db_name, model_name)
        data: dict[str, Any] = cast(dict[str, Any], await req.json())
        validation_schema = model.create_validation_schema()
        if validation_schema is not None:
            try:
                data = validation_schema.model_validate(data).model_dump()
            except ValidationError as exc:
                details = {}
                if hasattr(exc, "errors"):
                    for error in exc.errors():
                        details[error["loc"][0]] = error["msg"]
                return req.response.json({
                    "message": "Validation failed.",
                    "details": details
                }).status(HttpStatus.UNPROCESSABLE_ENTITY)
        db: SqlDatabase = self.dashboard.get_database(db_name)
        async with db.create_session() as session:
            async with session.begin():
                record = model()
                await record.admin_create(req, data, session)

        return req.res.json({
            "message": f"Record in {model_name} created successfully.",
            "status": "success"
        }).status(HttpStatus.CREATED)

    def generate_form(self, model: Type[DeclarativeBaseModel],
                      base_form: list[Any]) -> dict[str, FormField|Any]:
        """
        Generates the form for showing on form for modal
        """
        id_input_map: dict[str, FormField|Any] = {field.id: field for field in base_form}
        for field_id, field in model.custom_form_fields().items():
            id_input_map[field_id] = field
        ordering: Optional[list[str]] = model.form_fields_order()
        if ordering:
            ordered: dict[str, FormField|Any] = {}
            for field_id in ordering:
                ordered[field_id] = id_input_map[field_id]
            return ordered
        return id_input_map

    async def check_permission(self, perm_type: PermissionType,
                               req: Request,
                               db_name: str,
                               model_name: str) -> Type[DeclarativeBaseModel]:
        """Method for checking permissions for admin actions"""

        model: Optional[Type[DeclarativeBaseModel]] = self.dashboard.get_model(
            self.app._db_name_configs_map[db_name], model_name
        )
        if model is None:
            raise UnknownModelError(db_name, model_name)

        has_permission: bool = False
        if perm_type == PermissionType.CAN_VIEW:
            has_permission = await self.dashboard.has_view_permission(req, model)
        elif perm_type == PermissionType.CAN_CREATE:
            has_permission = await self.dashboard.has_create_permission(req, model)
        elif perm_type == PermissionType.CAN_UPDATE:
            has_permission = await self.dashboard.has_update_permission(req, model)
        else:
            has_permission = await self.dashboard.has_delete_permission(req, model)

        if not has_permission:
            raise AdminPermissionError("User does not have permission "
                                          f"to {perm_type} model data in {db_name} database.")
        return model
    
    def parse_key_value_path(self, s: str) -> dict[str, str]:
        """
        Parse a string like "id1/val1/id2/val2" into {"id1": "val1", "id2": "val2"}.
        Requires an even number of segments.
        """
        parts = s.split("/")
        if len(parts) % 2 != 0:
            raise ValueError(f"Invalid key/value path: {s!r}")

        return {parts[i]: parts[i + 1] for i in range(0, len(parts), 2)}
    
    def get_pk_filters_from_path(self, model: Type[DeclarativeBaseModel],
                                 model_pk_val: dict[str, str]) -> list:
        """Returns a list of filters"""
        filters: list = []
        for pk, val in model_pk_val.items():
            pk_attr = model.__table__.columns[pk]
            pk_val = pk_attr.type.python_type(val)
            filters.append(pk_attr == pk_val)
        return filters
    
    @staticmethod
    def create_attr_val_path_for_model(model, pk_names) -> str:
        """Creates url path for model with PKs and values"""
        path: str = ""
        for pk in pk_names:
            path+=f"/{pk}/{getattr(model, pk)}"
        return path[1:]



"""Run script for testg app"""

if __name__ == "__main__":
    import uvicorn
    from app.configs import Config
    ##Change parameters for starting the app (host, port etc)
    ##reload=True -> watches for file changes and reloads.
    configs = Config() #initilizes and makes defaults accessible
    uvicorn.run("app:Application", host=configs.HOST, port=configs.PORT,
                lifespan=configs.LIFESPAN, reload=configs.DEBUG, factory=True)



"""
MongoDB backend
"""
from typing import Optional, Callable, Any, Iterable, Mapping, TYPE_CHECKING, cast
from motor.motor_asyncio import AsyncIOMotorClient

from .async_nosql_backend_protocol import AsyncNoSqlBackendBase

if TYPE_CHECKING:
    from ....pyjolt import PyJolt

class MongoBackend(AsyncNoSqlBackendBase):
    def __init__(self, uri: str, database: Optional[str] = None) -> None:
        self._client: Optional[Any] = None
        self._db: Optional[Any] = None
        self._uri: str = uri
        self._database: Optional[str] = database
    
    @classmethod
    def configure_from_app(cls, app: "PyJolt", configs: dict[str, Any]) -> "AsyncNoSqlBackendBase":
        uri = cast(str, configs.get("DATABASE_URI"))
        database = cast(str, configs.get("DATABASE"))
        return cls(uri, database)

    async def connect(self) -> None:
        # Connect client
        self._client = AsyncIOMotorClient(self._uri)
        # Choose DB (or leave None; database_handle() will return client)
        self._db = self._client[self._database] if self._database else None

        #quick ping to verify connectivity
        try:
            await self._client.admin.command("ping")
        except Exception as exc:  # pragma: no cover
            raise RuntimeError(f"Failed to connect to MongoDB at {self._uri}: {exc}") from exc

    async def disconnect(self) -> None:
        if self._client is not None:
            self._client.close()
            self._client = None
            self._db = None

    def database_handle(self) -> Any:
        # Prefer a db handle if configured; otherwise, return client
        return self._db if self._db is not None else self._client

    def supports_transactions(self) -> bool:
        # Mongo supports transactions on replica sets / sharded clusters with WiredTiger.
        return True

    async def start_session(self) -> Any:
        if not self._client:
            raise RuntimeError("Mongo client not connected.")
        return await self._client.start_session()

    async def with_transaction(self, fn: Callable[..., Any], *args, session: Any = None, **kwargs) -> Any:
        # If session provided, run a transaction; else just call directly.
        if session is None:
            return await fn(*args, **kwargs)

        async def runner(sess):
            return await fn(*args, session=sess, **kwargs)

        # Motor session has method 'with_transaction'
        return await session.with_transaction(runner)

    # ---- Collection / CRUD ----

    def get_collection(self, name: str) -> Any:
        db_or_client = self.database_handle()
        if hasattr(db_or_client, "__getitem__"):  # db['collection']
            return db_or_client[name]
        # If only client exists but no db name given:
        raise RuntimeError("No database selected. Set NOSQL_DATABASE or use database['<db>'][collection].")

    async def find_one(self, collection: str, filter: Mapping[str, Any], *, session: Any = None, **kwargs) -> Any:
        coll = self.get_collection(collection)
        return await coll.find_one(filter, session=session, **kwargs)

    async def find_many(
        self,
        collection: str,
        filter: Mapping[str, Any] | None = None,
        *,
        session: Any = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        sort: Optional[Iterable[tuple[str, int]]] = None,
        **kwargs,
    ) -> list[Any]:
        filter = filter or {}
        coll = self.get_collection(collection)
        cursor = coll.find(filter, session=session, **kwargs)
        if sort:
            cursor = cursor.sort(list(sort))
        if skip:
            cursor = cursor.skip(skip)
        if limit:
            cursor = cursor.limit(limit)
        return await cursor.to_list(length=limit or 0)

    async def insert_one(self, collection: str, doc: Mapping[str, Any], *, session: Any = None, **kwargs) -> Any:
        coll = self.get_collection(collection)
        return await coll.insert_one(doc, session=session, **kwargs)

    async def insert_many(self, collection: str, docs: Iterable[Mapping[str, Any]], *, session: Any = None, **kwargs) -> Any:
        coll = self.get_collection(collection)
        return await coll.insert_many(list(docs), session=session, **kwargs)

    async def update_one(
        self, collection: str, filter: Mapping[str, Any], update: Mapping[str, Any], *,
        upsert: bool = False, session: Any = None, **kwargs
    ) -> Any:
        coll = self.get_collection(collection)
        return await coll.update_one(filter, update, upsert=upsert, session=session, **kwargs)

    async def delete_one(self, collection: str, filter: Mapping[str, Any], *, session: Any = None, **kwargs) -> Any:
        coll = self.get_collection(collection)
        return await coll.delete_one(filter, session=session, **kwargs)

    async def aggregate(
        self, collection: str, pipeline: Iterable[Mapping[str, Any]], *, session: Any = None, **kwargs
    ) -> list[Any]:
        coll = self.get_collection(collection)
        cursor = coll.aggregate(list(pipeline), session=session, **kwargs)
        return await cursor.to_list(length=None)

    async def execute_raw(self, *args, **kwargs) -> Any:
        """
        For MongoDB, you can call arbitrary db.command(...) via:
            await db.execute_raw("command", {"ping": 1})
        Or pass a callable to run with the database handle.
        """
        db = self.database_handle()
        if callable(args[0]):
            fn = args[0]
            return await fn(db, *args[1:], **kwargs)

        # Simple command passthrough:
        if isinstance(args[0], str):
            cmd = args[0]
            arg = args[1] if len(args) > 1 else {}
            return await db.command({cmd: arg} if not isinstance(arg, Mapping) else {cmd: 1, **arg})

        raise TypeError("Unsupported execute_raw usage for MongoBackend.")


"""
Post schemas
"""
from typing import List, Optional, Type, TYPE_CHECKING
from datetime import datetime
from sqlalchemy import or_, func, literal
from pydantic import BaseModel, field_validator, Field

from .base_schema import BaseResponse, QueryResponseSchema

if TYPE_CHECKING:
    from app.api.models.post import Post

def any_tag_in_csv_condition(tags: list[str], post: "Type[Post]"):
    """Creates a func for searching for tags"""
    # Build: ',' || lower(tags_list) || ',' LIKE '%,tag,%' OR ...
    csv = literal(",") + func.lower(post.tags_list) + literal(",")
    patterns = [f"%,{t.strip().lower()},%" for t in tags if t and t.strip()]
    return or_(*[csv.like(p) for p in patterns]) if patterns else literal(True)

class PostsQuery(BaseModel):
    """Query schema for posts"""
    user_id: Optional[int] = Field(None, description="Authors id")
    active: bool = Field(True, description="Status of post")
    created_at: datetime = Field(datetime.fromisoformat("2025-10-01"),
                                           description="Date of creation")
    tags: Optional[List[str]] = Field(None, description="Tags of the post")
    page: int = Field(1, description="Page of query results")
    per_page: int = Field(10, description="Results per page")


    @field_validator("tags", mode="before")
    @classmethod
    def split_comma_string(cls, v):
        """Transforms comma separated string to list"""
        if isinstance(v, str):
            # Split on commas, strip whitespace, and remove empty entries
            return [s.strip() for s in v.split(",") if s.strip()]
        return v

class PostUpdateInSchema(BaseModel):
    """Schema for updating post"""
    title_eng: Optional[str] = None
    title_slv: Optional[str] = None
    slug: Optional[str] = None
    content_eng: Optional[str] = None
    content_slv: Optional[str] = None
    tags: Optional[List[str]] = None

    @field_validator("tags", mode="before")
    @classmethod
    def split_comma_string(cls, v):
        """Transforms comma separated string to list"""
        if isinstance(v, str):
            # Split on commas, strip whitespace, and remove empty entries
            return [s.strip() for s in v.split(",") if s.strip()]
        return v

class PostInSchema(BaseModel):
    """Schema for creating post"""
    title_eng: str
    title_slv: str
    content_eng: str
    content_slv: str
    tags_list: List[str]

    @field_validator("tags_list", mode="before")
    @classmethod
    def split_comma_string(cls, v):
        """Transforms comma separated string to list"""
        if isinstance(v, str):
            # Split on commas, strip whitespace, and remove empty entries
            return [s.strip() for s in v.split(",") if s.strip()]
        return v


class PostOutSchema(BaseModel):
    """Posts query out schema"""
    id: int
    title_eng: str
    title_slv: str
    slug: str
    content_eng: str
    content_slv: str
    active: bool
    tags: List[str]
    author_id: int

    @classmethod
    def from_model(cls, post: "Post"):
        """Create schema from post model"""
        return cls(
            id=post.id,
            title_eng=post.title_eng,
            title_slv=post.title_slv,
            slug=post.slug,
            content_eng=post.content_eng,
            content_slv=post.content_slv,
            active=post.active,
            tags=post.tags,
            author_id=post.author_id
        )

class PostResponseSchema(BaseResponse):
    """Response schema for posts"""
    data: PostOutSchema

class PostsQueryDataSchema(QueryResponseSchema):
    """Posts API base query response"""
    items: list[PostOutSchema]

class PostsQueryResponseSchema(BaseResponse):
    """Posts API query response"""
    data: PostsQueryDataSchema



"""Publication data model"""
from datetime import datetime
from typing import TYPE_CHECKING, Any, Self, cast

from app.api.schemas.publication_schemas import PublicationOutSchema
from sqlalchemy import DateTime, Text, event
from sqlalchemy.orm import Mapped, mapped_column

from pyjolt.admin import register_model
from pyjolt.admin.input_fields import SelectInput
from pyjolt.database.sql import AsyncSession

from .base_model import DatabaseModel

if TYPE_CHECKING:
    from ..schemas.publication_schemas import PublicationsQuerySchema

@register_model
class Publication(DatabaseModel):
    """Post model"""
    __tablename__ = "publications"

    class Meta(DatabaseModel.Meta):
        exclude_from_create_form = ["created_at"]
        exclude_from_update_form = ["created_at"]
        exclude_from_table = ["funders_list", "abstract", "volume", "page", "created_at", "pub_type"]
        custom_labels = {
            "doi": "DOI",
            "authors": "Authors List",
            "container_title": "Journal Name/Book Title",
            "publisher": "Publisher",
            "id": "ID",
            "title": "Title"
        }
        custom_form_fields = [
            SelectInput(id="pub_type", name="pub_type", label="Publication type", options=[
                ("journal-article", "Journal Article"),
                ("book", "Book"),
                ("book-chapter", "Book Chapter")
            ])
        ]

    doi: Mapped[str] = mapped_column(nullable=False, unique=True)
    authors_list: Mapped[str] = mapped_column(nullable=False)
    title: Mapped[str] = mapped_column(nullable=False)
    pub_type: Mapped[str] = mapped_column(nullable=False, default="journal-article")
    publisher: Mapped[str] = mapped_column(nullable=True)
    container_title: Mapped[str] = mapped_column(nullable=False)
    page: Mapped[str] = mapped_column(nullable=True)
    volume: Mapped[str] = mapped_column(nullable=True)
    date_published: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    funders_list: Mapped[str] = mapped_column(nullable=True)
    abstract: Mapped[str] = mapped_column(Text, nullable=True)

    @property
    def authors(self) -> list[str]:
        """Authors as a list"""
        return self.authors_list.split(",")

    @authors.setter
    def authors(self, authors: list[str]):
        """Sets authors as a string"""
        self.authors_list = ",".join(authors)

    @property
    def funders(self) -> list[str]|None:
        """List of funders"""
        if self.funders_list is None:
            return None
        return self.funders_list.split(",")

    @funders.setter
    def funders(self, funders: list[str]):
        """Sets list of funders"""
        self.funders_list = ",".join(funders)

    @classmethod
    def from_crossref_api(cls, crossref_response: dict[str, Any]) -> Self:
        """Creates new publication instance from CrossrefApi data"""
        data: dict[str, Any] = cast(dict[str, Any], crossref_response.get("message"))
        authors: str = ",".join([f'{author["given"]} {author["family"]}'
                                 for author in data["author"]])
        #all_funders: Optional[list[dict[str, Any]]|str] = data.get("funder", None)
        #print(f'Authors for {data["DOI"]}: ', authors)
        return cls(
            doi=data["DOI"],
            authors_list=authors,
            title=data["title"][0],
            pub_type=data["type"],
            publisher=data["publisher"],
            container_title=data.get("container-title",[None])[0],
            page=data.get("page"),
            volume=data.get("volume"),
            date_published=datetime.fromisoformat(data["created"]["date-time"]),
            abstract=data.get("abstract")
        )

    @classmethod
    async def query_publications(cls, session: AsyncSession,
                           query_data: "PublicationsQuerySchema") -> dict[str, Any]:
        """Performs query"""
        conds = []
        if query_data.title is not None:
            conds.append(query_data.title in cls.title)
        if query_data.date_published is not None:
            conds.append(cls.date_published >= query_data.date_published)
        if query_data.publisher:
            conds.append(query_data.publisher in cls.publisher)
        if query_data.container_title is not None:
            conds.append(query_data.container_title in cls.container_title)
        if query_data.pub_type is not None:
            conds.append(cls.pub_type == query_data.pub_type)

        publications: dict[str, Any] = await cls.query(session).filter(
            *conds
        ).paginate(page=query_data.page, per_page=query_data.per_page)
        publications["items"] = [PublicationOutSchema.from_model(post)
                          for post in publications["items"]]
        return publications

@event.listens_for(Publication, "before_insert")
def set_date_published_before_insert(mapper, connection, target: Publication):
    if target.date_published and isinstance(target.date_published, str):
        target.date_published = datetime.fromisoformat(target.date_published)


@event.listens_for(Publication, "before_update")
def set_date_published_before_update(mapper, connection, target: Publication):
    if target.date_published and isinstance(target.date_published, str):
        target.date_published = datetime.fromisoformat(target.date_published)





"""
Router class for application routing. Uses Wrkzeug under the hood.
"""
from typing import Callable, Any, Mapping
from werkzeug.routing import Map, Rule
from werkzeug.exceptions import NotFound, MethodNotAllowed

class Router:
    """
    A Router class that leverages Werkzeug’s Map/Rule system.
    """
    def __init__(self, strict_slashes: bool = False):
        self.url_map = Map(strict_slashes=strict_slashes)
        # endpoint_name -> function
        self.endpoints: dict[str, Callable] = {}

    def add_route(self, path: str, endpoint: Callable, methods: list[str], endpoint_name: str):
        """
        Registers a route by creating a Rule in the internal Map.
        """
        if endpoint_name in self.endpoints:
            raise ValueError(f"Duplicate endpoint name registered: {endpoint_name}")
        # Store the endpoint function:
        self.endpoints[endpoint_name] = endpoint
        # Add a single Rule that handles the specified methods
        self.url_map.add(Rule(path, endpoint=endpoint_name, methods=methods))

    def match(self, path: str, method: str) -> tuple[Callable|None, Mapping[str, Any]]:
        """
        Matches the path and method against the routing map.
        Returns (endpoint_function, path_variables_dict) if found, otherwise (None, {}).
        """
        adapter = self.url_map.bind("", path_info=path)
        try:
            endpoint_name, kwargs = adapter.match(method=method)
            endpoint = self.endpoints.get(endpoint_name)
            return endpoint, kwargs
        except (NotFound, MethodNotAllowed) as exc:
            return None, {"method": method, "path": path, "exc": exc}




"""
CORS middleware for PyJolt.
Handles Cross-Origin Resource Sharing (CORS) by adding appropriate headers to HTTP responses.
"""
from typing import Any, Callable, Optional, TYPE_CHECKING, cast
from ..middleware import MiddlewareBase

if TYPE_CHECKING:
    from ..request import Request
    from ..response import Response

class CORSMiddleware(MiddlewareBase):
    """
    Middleware to handle CORS by adding necessary headers to HTTP responses.
    """

    async def middleware(self, req: "Request") -> "Response":
        scope = req.scope
        cors_opts = self._resolve_cors_options(cast(Callable, req.route_handler))
        origin = self._get_header(scope, "origin")
        is_preflight = (req.method == "OPTIONS") and bool(self._get_header(scope, "access-control-request-method"))

        if cors_opts["enabled"] and origin:
            if not self._origin_allowed(origin, cors_opts["allow_origins"]):
                return req.res.json({
                    "status":"error",
                    "message":"CORS origin not allowed"
                }).status(403)

            if is_preflight:
                return self._handle_preflight(scope, req, origin, cors_opts)

            # Wrap send to inject CORS headers on normal responses
            #pylint: disable-next=W0212
            req._send = self._wrap_cors_send(req._send, origin, cors_opts)

        return await self.next(req)

    def _get_header(self, scope, name: str) -> Optional[str]:
        target = name.lower().encode("latin1")
        for k, v in scope.get("headers", []):
            if k.lower() == target:
                try:
                    return v.decode("latin1")
                #pylint: disable-next=W0718
                except Exception:
                    return None
        return None

    def _normalize_list(self, v, *, upper: bool = False) -> list[str]:
        if v is None:
            return []
        if isinstance(v, str):
            v = [v]
        out = [s.strip() for s in v if s and s.strip()]
        if upper:
            out = [s.upper() for s in out]
        return out

    def _global_cors_options(self) -> dict[str, Any]:
        return {
            "enabled": self.app.get_conf("CORS_ENABLED"),
            "allow_origins": self.app.get_conf("CORS_ALLOW_ORIGINS"),
            "allow_methods": self.app.get_conf("CORS_ALLOW_METHODS"),
            "allow_headers": self.app.get_conf("CORS_ALLOW_HEADERS"),
            "expose_headers": self.app.get_conf("CORS_EXPOSE_HEADERS"),
            "allow_credentials": self.app.get_conf("CORS_ALLOW_CREDENTIALS"),
            "max_age": self.app.get_conf("CORS_MAX_AGE"),
        }

    def _resolve_cors_options(self, handler: Optional[Callable]) -> dict[str, Any]:
        """
        Merge global and endpoint-level CORS options.
        If handler has @no_cors, CORS is completely disabled.
        """
        opts = self._global_cors_options()

        if handler is None:
            return opts

        # Check if handler explicitly disables CORS
        if getattr(handler, "_disable_cors", False):
            opts["enabled"] = False
            return opts

        # Merge @cors decorator overrides
        overrides = getattr(handler, "_cors_options", None) or {}
        for k, v in overrides.items():
            if v is not None:
                opts[k] = v

        opts["allow_origins"] = self._normalize_list(opts["allow_origins"])
        opts["allow_methods"] = self._normalize_list(opts["allow_methods"], upper=True)
        opts["allow_headers"] = self._normalize_list(opts["allow_headers"])
        opts["expose_headers"] = self._normalize_list(opts["expose_headers"])
        return opts

    def _origin_allowed(self, origin: Optional[str], allow_origins: list[str]) -> bool:
        if not origin:
            return True
        if "*" in allow_origins:
            return True
        return origin in allow_origins

    def _build_cors_headers_for_preflight(
        self,
        *,
        origin: str,
        request_method: Optional[str],
        request_headers: Optional[str],
        opts: dict[str, Any]
    ) -> dict[str, str]:
        allow_methods = opts["allow_methods"]
        allow_headers = opts["allow_headers"]

        # Echo requested headers if none configured
        if not allow_headers and request_headers:
            allow_headers = [h.strip() for h in request_headers.split(",") if h.strip()]

        # If "*" configured but credentials allowed: echo back the Origin
        allow_origin_value = "*"
        if opts["allow_credentials"] or ("*" not in opts["allow_origins"]):
            allow_origin_value = origin

        headers: dict[str, str] = {
            "access-control-allow-origin": allow_origin_value,
            "vary": "Origin",
            "access-control-allow-methods": ", ".join(allow_methods),
            "access-control-allow-headers": ", ".join(allow_headers),
        }

        if opts["allow_credentials"]:
            headers["access-control-allow-credentials"] = "true"

        if opts.get("max_age") is not None:
            headers["access-control-max-age"] = str(int(opts["max_age"]))

        return headers

    def _build_cors_headers_for_response(
        self,
        *,
        origin: Optional[str],
        opts: dict[str, Any]
    ) -> list[tuple[bytes, bytes]]:
        if not origin:
            return []

        # If "*" configured but credentials allowed: echo Origin instead.
        allow_origin_value = "*"
        if opts["allow_credentials"] or ("*" not in opts["allow_origins"]):
            allow_origin_value = origin

        headers: list[tuple[bytes, bytes]] = [
            (b"access-control-allow-origin", allow_origin_value.encode("latin1")),
            (b"vary", b"Origin"),
        ]
        if opts["allow_credentials"]:
            headers.append((b"access-control-allow-credentials", b"true"))
        expose = opts.get("expose_headers") or []
        if expose:
            headers.append((b"access-control-expose-headers", ", ".join(expose).encode("latin1")))
        return headers

    def _handle_preflight(self, scope, req: "Request", origin: str, cors_opts: dict) -> "Response":
        acr_method = self._get_header(scope, "access-control-request-method")
        acr_headers = self._get_header(scope, "access-control-request-headers")

        if(acr_method and cors_opts["allow_methods"] 
           and acr_method.upper() not in cors_opts["allow_methods"]):
            return req.res.text("Preflight method not allowed").status(405)

        headers = self._build_cors_headers_for_preflight(
            origin=origin,
            request_method=acr_method,
            request_headers=acr_headers,
            opts=cors_opts,
        )
        return req.res.set_headers(headers)

    def _wrap_cors_send(self, send, origin: str, cors_opts: dict):
        cors_headers = self._build_cors_headers_for_response(origin=origin, opts=cors_opts)
        async def cors_send(event):
            if event.get("type") == "http.response.start":
                headers = event.get("headers", [])
                event = {**event, "headers": headers + cors_headers}
            await send(event)
        return cors_send


"""
Publication schemas
"""
from datetime import datetime
from typing import Optional, Self, TYPE_CHECKING
from pydantic import BaseModel, Field, field_serializer
from .base_schema import BaseResponse, QueryResponseSchema

if TYPE_CHECKING:
    from ..models.publication import Publication

class PublicationsQuerySchema(BaseModel):
    """Schema for publication querying"""
    title: Optional[str] = Field(None, description="Title or part of title")
    pub_type: Optional[str] = Field(None, description="Type of publication")
    publisher: Optional[str] = Field(None, description="Publication publisher")
    container_title: Optional[str] = Field(None, description="Journal/Book title")
    date_published: datetime = Field(datetime.fromisoformat("2000-01-01"),
                                description="Date of publication")
    page: int = Field(1, description="Page of query results")
    per_page: int = Field(10, description="Results per page")

class PublicationOutSchema(BaseModel):
    """Publication out schema"""
    doi: str
    authors: list[str]
    title: str
    pub_type: str
    publisher: str
    container_title: str
    page: Optional[str] = None
    volume: Optional[str] = None
    date_published: Optional[datetime] = None
    abstract: Optional[str] = None

    @field_serializer('date_published')
    def serialize_date_published(self, dt: datetime, _info):
        """Serialized date_published to ISO string"""
        return dt.strftime('%Y-%m-%d %H:%M:%S')

    @classmethod
    def from_model(cls, model: "Publication") -> Self:
        """Creates schema from model"""
        return cls(
            doi=model.doi,
            authors=model.authors,
            title=model.title,
            pub_type=model.pub_type,
            publisher=model.publisher,
            container_title=model.container_title,
            page=model.page,
            volume=model.volume,
            date_published=model.date_published,
            abstract=model.abstract
        )

class PublicationResponseSchema(BaseResponse):
    """Publications API response schema"""
    data: PublicationOutSchema

class PublicationsQueryDataSchema(QueryResponseSchema):
    """Publication API base query response"""
    items: list[PublicationOutSchema]

class PublicationsQueryResponseSchema(BaseResponse):
    """Publications API query response"""
    data: PublicationsQueryDataSchema



"""
Controller decorators
"""

from typing import (
    Any,
    Awaitable,
    Callable,
    Optional,
    ParamSpec,
    Protocol,
    Type,
    cast,
)
from functools import wraps
import inspect

from .controller import Controller, Descriptor
from .utilities import (
    _extract_response_type,
    get_type_hints,
    _unwrap_annotated,
    _is_pydantic_model,
    _content_type_matches,
    _read_payload_for_consumes,
    _build_model,
)
from ..response import Response
from ..request import Request
from ..utilities import run_sync_or_async
from ..exceptions import MethodNotControllerMethod, UnexpectedDecorator
from ..media_types import MediaType
from ..http_methods import HttpMethod
from ..http_statuses import HttpStatus

P = ParamSpec("P")
R = Any  # Return type of the *original* endpoint method

# Any bound method (sync or async) -> async method returning Response
AsyncMethod = Callable[..., Awaitable["Response"]]

#pylint: disable-next=R0903
class _EndpointDecorator(Protocol):
    """A decorator that accepts a sync or async method and returns an async method."""
    def __call__(self, func: Callable[..., Any]) -> AsyncMethod: ...

#pylint: disable-next=R0903
class EndpointDecoratorFn(Protocol):
    """Callable returned by HTTP method factories (post/put/patch/delete)."""
    def __call__(
        self,
        url_path: str,
        open_api_spec: bool = True,
        tags: Optional[list[str]] = None,
    ) -> _EndpointDecorator: ...

def get(
    url_path: str, open_api_spec: bool = True, tags: Optional[list[str]] = None
) -> _EndpointDecorator:
    """GET http handler decorator."""

    def decorator(func: Callable[..., Any]) -> AsyncMethod:
        @wraps(func)
        async def wrapper(self: Controller, *args: Any, **kwargs: Any) -> "Response":
            if not isinstance(self, Controller):
                raise MethodNotControllerMethod(
                    f"Method {func.__name__} is not part of a valid controller class"
                )

            # pre-hooks
            req: "Request" = args[0]  # type: ignore[index]
            for m in reversed(getattr(self, "_controller_decorator_methods", []) or []):
                await run_sync_or_async(m, req)
            for m in getattr(self, "_before_request_methods", []) or []:
                await run_sync_or_async(m, req)

            # call the original (sync or async)
            response: "Response" = await run_sync_or_async(func, self, *args, **kwargs)

            # post-hooks
            for m in getattr(self, "_after_request_methods", []) or []:
                await run_sync_or_async(m, response)
            return response

        merged = {
            **(getattr(func, "_handler", {}) or {}),
            "http_method": HttpMethod.GET.value,
            "path": url_path,
            "open_api_spec": open_api_spec,
            "tags": tags if tags is not None else [],
        }
        if merged.get("consumes", False):
            raise UnexpectedDecorator("GET endpoints can't consume request bodies.")
        # pylint: disable=protected-access
        wrapper._handler = merged  # type: ignore[attr-defined]
        return wrapper

    return cast(_EndpointDecorator, decorator)

def endpoint_decorator_factory(http_method: HttpMethod) -> EndpointDecoratorFn:
    def endpoint_decorator(
        url_path: str,
        open_api_spec: bool = True,
        tags: Optional[list[str]] = None,
    ) -> _EndpointDecorator:
        def decorator(func: Callable[..., Any]) -> AsyncMethod:
            @wraps(func)
            async def wrapper(self: Controller, *args: Any, **kwargs: Any) -> "Response":
                if not isinstance(self, Controller):
                    raise MethodNotControllerMethod(
                        f"Method {func.__name__} is not part of a valid controller class"
                    )

                # Optionally infer/record expected response type for endpoint
                req: "Request" = args[0]  # type: ignore[index]
                if req.response.expected_body_type is None:
                    expected = _extract_response_type(func)
                    # pylint: disable=protected-access
                    req.response._set_expected_body_type(expected)

                # pre-hooks
                for m in reversed(getattr(self, "_controller_decorator_methods", []) or []):
                    await run_sync_or_async(m, req)
                for m in getattr(self, "_before_request_methods", []) or []:
                    await run_sync_or_async(m, req)

                # call the original (sync or async)
                response: "Response" = await run_sync_or_async(func, self, *args, **kwargs)

                # post-hooks
                for m in getattr(self, "_after_request_methods", []) or []:
                    await run_sync_or_async(m, response)
                return response

            # attach metadata
            # pylint: disable=protected-access
            wrapper._handler = {  # type: ignore[attr-defined]
                **(getattr(func, "_handler", {}) or {}),
                "http_method": http_method.value,
                "path": url_path,
                "open_api_spec": open_api_spec,
                "tags": tags if tags is not None else [],
            }
            return wrapper

        return cast(_EndpointDecorator, decorator)

    return endpoint_decorator


post = endpoint_decorator_factory(HttpMethod.POST)
put = endpoint_decorator_factory(HttpMethod.PUT)
patch = endpoint_decorator_factory(HttpMethod.PATCH)
delete = endpoint_decorator_factory(HttpMethod.DELETE)

def socket(
    url_path: str) -> _EndpointDecorator:
    """SOCKET http handler decorator."""

    def decorator(func: Callable[..., Any]) -> AsyncMethod:
        @wraps(func)
        async def wrapper(self: Controller, *args: Any, **kwargs: Any):
            if not isinstance(self, Controller):
                raise MethodNotControllerMethod(
                    f"Method {func.__name__} is not part of a valid controller class"
                )
            await run_sync_or_async(func, self, *args, **kwargs)

        merged = {
            **(getattr(func, "_handler", {}) or {}),
            "http_method": HttpMethod.SOCKET.value,
            "path": url_path,
            "open_api_spec": False,
            "tags": None,
        }
        # pylint: disable=protected-access
        wrapper._handler = merged  # type: ignore[attr-defined]
        return wrapper

    return cast(_EndpointDecorator, decorator)

def consumes(media_type: MediaType) -> _EndpointDecorator:
    """Decorator indicating what media type the endpoint consumes."""

    def decorator(func: Callable[..., Any]) -> AsyncMethod:
        sig = inspect.signature(func)

        try:
            hints = get_type_hints(func, include_extras=True)
        # pylint: disable-next=W0718
        except Exception:
            hints = getattr(func, "__annotations__", {}) or {}

        consumed_type: Optional[Any] = None
        for name, param in list(sig.parameters.items())[2:]:
            ann = hints.get(name, param.annotation)
            ann = _unwrap_annotated(ann)
            if _is_pydantic_model(ann):
                consumed_type = ann
                break

        @wraps(func)
        async def wrapper(self: Controller, *args: Any, **kwargs: Any) -> "Response":
            # Request is auto-injected as the first arg after self
            if not args:
                raise RuntimeError(
                    "Request must be auto-injected as the first argument after self."
                )
            req: "Request" = args[0]  # type: ignore[index]

            incoming_ct = req.headers.get("content-type", "")
            if not _content_type_matches(incoming_ct, media_type):
                return req.response.json(
                    {
                        "detail": "Unsupported Media Type",
                        "expected": media_type.value,
                        "received": incoming_ct or None,
                    }
                ).status(415)

            payload = await _read_payload_for_consumes(req, media_type)
            # Parameters: [0]=self, [1]=req, others start at index 2
            for name, param in list(sig.parameters.items())[2:]:
                if name in kwargs:
                    continue

                ann = hints.get(name, param.annotation)
                if _is_pydantic_model(ann):
                    kwargs[name] = _build_model(ann, payload)

            return await run_sync_or_async(func, self, *args, **kwargs)

        prev = getattr(func, "_handler", {}) or {}
        merged = dict(prev)
        merged.update({"consumes": media_type, "consumes_type": consumed_type})
        # pylint: disable=protected-access
        wrapper._handler = merged  # type: ignore[attr-defined]
        return wrapper

    return decorator


def produces(
    media_type: MediaType, status_code: HttpStatus = HttpStatus.OK
) -> _EndpointDecorator:
    """
    Decorator indicating what media types the endpoint 
    produces and what the default status code is.
    """

    def decorator(func: Callable[..., Any]) -> AsyncMethod:
        expected_body = _extract_response_type(func)

        @wraps(func)
        async def wrapper(self: Controller, *args: Any, **kwargs: Any) -> "Response":
            # Request is auto-injected as first arg after self
            if not args:
                raise RuntimeError(
                    "Request must be auto-injected as the first argument after self."
                )
            req: "Request" = args[0]  # type: ignore[index]
            # pylint: disable=protected-access
            req.response._set_expected_body_type(expected_body)

            res: "Response" = await run_sync_or_async(func, self, *args, **kwargs)
            res.set_header("content-type", media_type.value)
            if status_code != HttpStatus.OK:
                res.status(status_code)
            return res

        # Preserve/merge handler metadata (produces list)
        prev = getattr(func, "_handler", {}) or {}
        merged = dict(prev)
        merged.update({"produces": media_type, "default_status_code": status_code})
        # pylint: disable=protected-access
        wrapper._handler = merged  # type: ignore[attr-defined]
        return wrapper

    return decorator

def open_api_docs(*args: Descriptor):
    """Adds descriptions for error responses to OpenAPI documentation"""

    def decorator(func: AsyncMethod) -> AsyncMethod:
        prev = getattr(func, "_handler", {}) or {}
        merged = dict(prev)
        merged.update({"error_responses": list(args)})
        # pylint: disable=protected-access
        func._handler = merged  # type: ignore[attr-defined]
        return func

    return decorator

def before_request(func: Callable[..., Any]) -> Callable[..., Any]:
    setattr(func, "_before_request", True)
    return func

def after_request(func: Callable[..., Any]) -> Callable[..., Any]:
    setattr(func, "_after_request", True)
    return func

def cors(
    *,
    allow_origins: Optional[list[str]] = None,
    allow_methods: Optional[list[str]] = None,
    allow_headers: Optional[list[str]] = None,
    expose_headers: Optional[list[str]] = None,
    allow_credentials: Optional[bool] = None,
    max_age: Optional[int] = None,
) -> Callable:
    """
    Per-endpoint CORS override. Any provided option overrides the global config.
    Usage:
    ```
        @path("/items")
        @cors(allow_origins=["https://app.example.com"], allow_credentials=True)
        async def list_items(self, req: Request) -> Response: ...
    ```
    """
    def decorator(func: Callable) -> Callable:
        setattr(func, "_cors_options", {
            "allow_origins": allow_origins,
            "allow_methods": allow_methods,
            "allow_headers": allow_headers,
            "expose_headers": expose_headers,
            "allow_credentials": allow_credentials,
            "max_age": max_age,
        })
        return func
    return decorator

def no_cors(func: Callable) -> Callable:
    """
    Decorator to disable CORS for a specific endpoint handler.
    Usage:
    ```
        @path("/internal")
        @no_cors
        async def internal(self, req: Request) -> Response:
            return req.res.text("No CORS here")
    ```
    """
    setattr(func, "_disable_cors", True)
    return func

def development(func_or_cls: Callable|Type[Any]) -> Callable|Type:
    """
    Decorator to mark a controller or endpoint as development only.
    The decorated controller or endpoint will be unreachable if the
    application is not in DEBUG mode.
    """
    setattr(func_or_cls, "_development", True)
    return func_or_cls



"""
Posts api
"""
from typing import Any, Optional, cast
from pyjolt import MediaType, Response, Request, HttpStatus, abort
from pyjolt.controller import (Controller, path, get, post, put, produces,
                               delete, open_api_docs, Descriptor,
                               consumes)
from pyjolt.database.sql import AsyncSession
from pyjolt.auth import login_required

from app.extensions import db
from app.api.models.post import Post
from app.api.schemas.base_schema import ErrorResponseSchema
from app.api.schemas.post_schemas import (PostsQuery,
                                          PostOutSchema,
                                          PostInSchema,
                                          PostUpdateInSchema,
                                          PostResponseSchema,
                                          PostsQueryResponseSchema)

@path("/api/v1/posts")
class PostsApi(Controller):
    """Posts controller"""

    @get("/")
    @db.managed_session
    @produces(MediaType.APPLICATION_JSON)
    async def get_posts(self, req: Request,
                session: AsyncSession) -> Response[PostsQueryResponseSchema]:
        """Returns all posts"""
        query: PostsQuery = PostsQuery.model_validate(
                                            req.query_params)
        results: dict[str, Any] = await Post.query_posts(session, query)
        return req.res.json({
            "message": "Query successful.",
            "status": "success",
            "data": results
        }).status(HttpStatus.OK)

    @get("/<int:post_id>")
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND,
                              description="Post not found",
                              body=ErrorResponseSchema))
    @db.managed_session
    async def get_post(self, req: Request, post_id: int,
                       session: AsyncSession) -> Response[PostResponseSchema]:
        """Gets post by id"""
        post_data: Optional[Post] = await Post.query(
            session).filter_by(id=post_id).first()
        if post_data is None:
            abort(f"Post with id {post_id} not found.",
                  status_code=HttpStatus.NOT_FOUND)

        return req.res.json({
            "message": "Post fetched successful",
            "status": "success",
            "data": PostOutSchema.from_model(cast(Post,post_data))
        }).status(HttpStatus.OK)

    @post("/")
    @produces(MediaType.APPLICATION_JSON, status_code=HttpStatus.CREATED)
    @consumes(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.UNAUTHORIZED,
                              description="Not authorized",
                              body=ErrorResponseSchema))
    @db.managed_session
    @login_required
    async def create_post(self, req: Request,
                          post_data: PostInSchema,
                          session: AsyncSession) -> Response[PostResponseSchema]:
        """Creates a new post"""
        new_post: Post = Post(**post_data.model_dump())
        session.add(new_post)
        await session.flush() #to get the ID of the post
        return req.res.json({
            "message": "Post created successfully.",
            "status": "success",
            "data": PostOutSchema.from_model(new_post)
        }).status(HttpStatus.CREATED)

    @delete("/<int:post_id>")
    @produces(MediaType.NO_CONTENT, status_code=HttpStatus.NO_CONTENT)
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND, description="Post not found",
                              body=ErrorResponseSchema),
                Descriptor(status=HttpStatus.UNAUTHORIZED, description="Not authorized",
                              body=ErrorResponseSchema))
    @db.managed_session
    @login_required
    async def delete_post(self, req: Request, post_id: int,
                          session: AsyncSession) -> Response:
        """Deletes post by id"""
        post_data: Post = await Post.query(session).filter_by(id=post_id).first()
        if post_id is None:
            abort(f"Post with id {post_id} not found",
                  status_code=HttpStatus.NOT_FOUND)

        await session.delete(post_data)
        return req.res.no_content()

    @put("/<int:post_id>")
    @consumes(MediaType.APPLICATION_JSON)
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND, description="Post not found",
                              body=ErrorResponseSchema),
                   Descriptor(status=HttpStatus.UNAUTHORIZED,
                              description="You are not authorized for this action.",
                              body=ErrorResponseSchema))
    @db.managed_session
    @login_required
    async def update_post(self, req: Request, post_id: int,
                          session: AsyncSession,
                          post_data: PostUpdateInSchema) -> Response[PostResponseSchema]:
        """Updates post with post_id"""
        org_post: Post = await Post.query(session).filter_by(id=post_id).first()
        if org_post is None:
            abort(f"Post with id {post_id} not found.", status_code=HttpStatus.NOT_FOUND)

        if org_post.author_id != req.user.id:
            abort("You are not authorized for this action.", status_code=HttpStatus.UNAUTHORIZED)

        for key, value in post_data.model_dump().items():
            if value is not None:
                setattr(org_post, key, value)
        session.add(org_post)

        return req.res.json({
            "message": "Post updated successfully",
            "status": "success",
            "data": PostOutSchema.from_model(org_post)
        }).status(HttpStatus.OK)



"""
Email client extension
"""
import aiosmtplib as smtplib
from email.message import EmailMessage
from typing import (TYPE_CHECKING, Optional, cast,
                    TypedDict, NotRequired)
from pydantic import BaseModel, Field
from jinja2 import Environment

from ..base_extension import BaseExtension
from ..utilities import run_sync_or_async

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

class _EmailConfigs(BaseModel):
    """
    Email client configuration model
    """
    SENDER_NAME_OR_ADDRESS: str = Field(description="The name or address of the email sender")
    SMTP_SERVER: str = Field(description="SMTP server address")
    SMTP_PORT: int = Field(description="SMTP server port")
    USERNAME: Optional[str] = Field(None, description="SMTP username")
    PASSWORD: Optional[str] = Field(None, description="SMTP password")
    USE_TLS: Optional[bool] = Field(False, description="Use TLS for SMTP connection")

class EmailConfig(TypedDict):
    """
    Email client configuration dictionary
    """
    SENDER_NAME_OR_ADDRESS: str
    SMTP_SERVER: str
    SMTP_PORT: int
    USERNAME: NotRequired[str]
    PASSWORD: NotRequired[str]
    USE_TLS: NotRequired[bool]

class EmailClient(BaseExtension):
    """
    Email client extension class
    """

    def __init__(self, configs_name: str = "EMAIL_CLIENT"):
        self._app: "Optional[PyJolt]" = None
        self._configs_name = configs_name
        self._configs: dict[str, str|int|bool] = {}
        self.render_engine: Environment = None  # type: ignore

    def init_app(self, app: "PyJolt") -> None:
        """Initilizes the extension with the PyJolt app"""
        self._app = app
        self._configs = app.get_conf(self._configs_name, {})
        self._configs = self.validate_configs(self._configs, _EmailConfigs)

        self._app.add_extension(self)
        self.render_engine = self._app.jinja_environment

    def get_client(self) -> smtplib.SMTP:
        """Returns the email client instance"""
        return smtplib.SMTP(
            hostname=cast(str,self._configs.get("SMTP_SERVER")),
            port=cast(int,self._configs.get("SMTP_PORT"))
        )
    
    async def send_email_with_template(self, to_address: str|list[str],
                                subject: str, template_path: str,
                                attachments: Optional[dict[str, bytes]] = None,
                                context: Optional[dict] = None) -> None:
        """Sends an email using a template and context data"""
        
        if context is None:
            context = {}

        for method in self.app.global_context_methods:
            additional_context = await run_sync_or_async(method)
            if not isinstance(additional_context, dict):
                raise ValueError("Return of global context method must be off type dictionary")
            context = {**context, **additional_context}
        context["url_for"] = self.app.url_for

        template = self.render_engine.get_template(template_path)
        rendered = await template.render_async(**context)
        await self.send_email(to_address, subject, rendered, attachments)
    
    async def send_email(
        self,
        to_address: str|list[str],
        subject: str,
        body: str,
        attachments: Optional[dict[str, bytes]] = None
    ) -> None:
        """
        Sends an email with attachments using the SMTP client

        :param to_address: Recipient email address or comma separated list of addresses
        :param subject: Email subject
        :param body: Email body
        :param attachments: Dictionary of attachment filenames and their byte content
        """
        if isinstance(to_address, str):
            to_address = [to_address]

        msg: EmailMessage = EmailMessage()
        msg['From'] = cast(str,self._configs.get("SENDER_NAME_OR_ADDRESS"))
        msg['To'] = ', '.join(to_address)
        msg['Subject'] = subject
        msg.set_content(body)

        if attachments:
            for filename, filecontent in attachments.items():
                msg.add_attachment(filecontent, maintype='application', subtype='octet-stream', filename=filename)

        async with self.get_client() as client:
            if cast(bool,self._configs.get("USE_TLS")):
                await client.starttls()
            if self._configs.get("USERNAME") is not None and self._configs.get("PASSWORD") is not None:
                await client.login(
                    cast(str,self._configs.get("USERNAME")),
                    cast(str,self._configs.get("PASSWORD"))
                )
            await client.send_message(msg)



"""
Middleware base class
"""
from abc import abstractmethod, ABC
from typing import Callable, TYPE_CHECKING, Awaitable, Any, Protocol

from pydantic import BaseModel, ValidationError

from .utilities import run_sync_or_async

if TYPE_CHECKING:
    from .pyjolt import PyJolt
    from .request import Request
    from .response import Response

# One request in -> Response out (async)
class AppCallableType(Protocol):
    def __call__(self, req: "Request") -> Awaitable["Response"]: ...

# A middleware factory: given (app_instance, next_app) returns a wrapped app
MiddlewareFactory = Callable[["PyJolt", AppCallableType], AppCallableType]

class MiddlewareBase(ABC):
    """
    Base class for middleware
    """
    configs_name: str

    def __init__(self, app: "PyJolt", next_app: AppCallableType):
        """
        Accepts the application and the next part of the middleware chain
        """
        self._app = app
        self._next = next_app
    
    def validate_configs(self, configs: dict[str, Any], model: type[BaseModel]) -> dict[str, Any]:
        try:
            return model.model_validate(configs).model_dump()
        except ValidationError as e:
            raise ValueError(f"Invalid configuration for {self.configs_name or self.__class__.__name__}: {e}") from e

    @abstractmethod
    async def middleware(self, req: "Request") -> "Response":
        """
        Middleware method to be implemented by subclasses
        """
        ...

    async def __call__(self, req: "Request") -> "Response":
        """
        Middleware call method
        """
        return await run_sync_or_async(self.middleware, req)

    @property
    def app(self) -> "PyJolt":
        """
        Returns the application instance
        """
        return self._app
    
    @property
    def next(self) -> "Callable[[Request], Awaitable[Any]]":
        """
        Returns the next part of the middleware chain
        """
        return self._next



# base_protocol.py
#pylint: disable=W0613

from __future__ import annotations
from typing import Any, Optional, Tuple, cast, Type, Protocol, TYPE_CHECKING
from pydantic import BaseModel

from sqlalchemy import Column
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.ext.asyncio import AsyncSession
from .sqlalchemy_async_query import AsyncQuery

if TYPE_CHECKING:
    from ...admin.input_fields import FormField

from ...request import Request

class MetaProtocol(Protocol):

    exclude_from_create_form: list[str]
    exclude_from_update_form: list[str]
    exclude_from_table: list[str]

    add_to_form: dict[str, FormField]

    custom_labels: dict[str, str]
    custom_form_fields: list[FormField]

    form_fields_order: list[str]

    create_validation_shema: Type[BaseModel]
    update_validation_shema: Type[BaseModel]

class DeclarativeBaseModel(DeclarativeBase):
    """
    Defines the interface that the custom
    DeclarativeBase class must satisfy.
    """
    __db_name__: str
    __db_configs_name__: str
    __abstract__ = True

    class Meta(MetaProtocol):
        exclude_from_create_form: list[str]
        exclude_from_update_form: list[str]
        exclude_from_table: list[str]

        add_to_form: dict[str, FormField]

        custom_labels: dict[str, str]
        custom_form_fields: list[FormField]

        form_fields_order: list[str]

        create_validation_shema: Type[BaseModel]
        update_validation_shema: Type[BaseModel]

    def __init__(self, **kwargs):
        if kwargs is not None:
            for key, value in kwargs.items():
                setattr(self, key, value)
    
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not cls.__abstract__:
            if "__db_name__" not in cls.__dict__:
                raise TypeError(
                    f"{cls.__name__} must define a class attribute '__db_name__'"
                )

    async def admin_create(self, req: "Request",
                           new_data: dict[str, Any],
                           session: "AsyncSession") -> None:
        """
        Saves the current instance to the database. Used in admin dashboard forms
        for creating and editing records. If customization is needed, override this method
        in the model class.
        """
        for key, value in new_data.items():
            setattr(self, key, value)
        session.add(self)

    async def admin_delete(self, req: "Request",
                           session: "AsyncSession") -> None:
        """
        Deletes the current instance from the database. Used in admin dashboard
        for deleting records. If customization is needed, override this method
        in the model class.
        """
        await session.delete(self)

    async def admin_update(self, req: "Request",
                           new_data: dict[str, Any],
                           session: "AsyncSession") -> None:
        """
        Updates the current instance with new data. Used in admin dashboard
        forms for editing records. If customization is needed, override this method
        in the model class.
        """
        for key, value in new_data.items():
            setattr(self, key, value)
        session.add(self)

    @classmethod
    def query(cls, session: AsyncSession) -> AsyncQuery:
        return AsyncQuery(session, cls)

    @classmethod
    def db_name(cls) -> str:
        return cls.__db_name__
    
    @classmethod
    def db_configs_name(cls) -> str:
        return cls.__db_configs_name__
    
    @classmethod
    def primary_key_names(cls) -> Optional[list[str]]:
        """
        Returns the attribute names of the primary key columns.
        """
        mapper = inspect(cls)
        pks = mapper.primary_key

        if not pks:
            return None

        return [pk.key for pk in pks]#pks[0].key
    
    @classmethod
    def primary_keys(cls) -> Optional[Tuple[Column[Any]]]:
        mapper = inspect(cls)
        pks = mapper.primary_key
        if not pks:
            return None
        return cast(Tuple[Column[Any]], pks)
    
    @classmethod
    def exclude_from_create_form(cls) -> list[str]:
        """Returns all fields that are declared as hidden in the form"""
        if not hasattr(cls.Meta, "exclude_from_create_form"):
            return []
        return cls.Meta.exclude_from_create_form
    
    @classmethod
    def exclude_from_update_form(cls) -> list[str]:
        """Returns all fields that are declared as hidden in the form"""
        if not hasattr(cls.Meta, "exclude_from_update_form"):
            return []
        return cls.Meta.exclude_from_update_form
    
    @classmethod
    def exclude_from_table(cls) -> list[str]:
        """Returns all fields that are declared as excluded in the table"""
        if not hasattr(cls.Meta, "exclude_from_table"):
            return []
        return cls.Meta.exclude_from_table
    
    @classmethod
    def form_labels_map(cls) -> dict[str, str]:
        """Map of attribute names -> human readable names"""
        if not hasattr(cls.Meta, "custom_labels"):
            return {}
        return cls.Meta.custom_labels
    
    @classmethod
    def custom_form_fields(cls) -> dict[str, FormField]:
        """Returns custom form fields for the admin dashboard forms"""
        if not hasattr(cls.Meta, "custom_form_fields"):
            return {}
        return {input.id: input for input in cls.Meta.custom_form_fields}
    
    @classmethod
    def add_to_form(cls) -> dict[str, Any]:
        """Returns additional form fields to be added to admin dashboard forms"""
        if not hasattr(cls.Meta, "add_to_form"):
            return {}
        return cls.Meta.add_to_form
    
    @classmethod
    def create_validation_schema(cls) -> Optional[Type[BaseModel]]:
        """Returns the schema used for validating creation forms in admin dashboard"""
        if not hasattr(cls.Meta, "create_validation_shema"):
            return None
        return cls.Meta.create_validation_shema
    
    @classmethod
    def update_validation_schema(cls) -> Optional[Type[BaseModel]]:
        """Returns the schema used for validating edit forms in admin dashboard"""
        if not hasattr(cls.Meta, "update_validation_shema"):
            return None
        return cls.Meta.update_validation_shema
    
    @classmethod
    def order_table_by(cls) -> Optional[list[str]]:
        """Returns list of strings used for ordering the table in admin dashboard"""
        if not hasattr(cls.Meta, "order_table_by"):
            return None
        return cls.Meta.order_table_by
    
    @classmethod
    def form_fields_order(cls) -> Optional[list[str]]:
        """
        Returns a list of string with the order of input fields or None
        """
        if not hasattr(cls.Meta, "form_fields_order"):
            return None
        return cls.Meta.form_fields_order






"""Researchers API"""
from pyjolt import HttpStatus, MediaType, Request, Response, abort
from pyjolt.controller import (Controller, path,
                               get, produces,
                               open_api_docs, Descriptor)
from pyjolt.database.sql import AsyncSession

from app.extensions import db
from app.api.models.researcher import Researcher
from app.api.schemas.base_schema import ErrorResponseSchema
from app.api.schemas.researcher_schemas import (ResearcherOutSchema,
                                                AllResearchersResponseSchema,
                                                ResearcherResponseSchema)

@path("/api/v1/researchers")
class ResearchersApi(Controller):
    """Researchers api controller"""

    @get("/")
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    async def get_researchers(self, req: Request,
                        session: AsyncSession) -> Response[AllResearchersResponseSchema]:
        """Fetches all researchers"""
        researchers: list[Researcher] = await Researcher.query(session).all()

        return req.res.json({
            "message": "Researchers fetched successfully",
            "status": "success",
            "data": [ResearcherOutSchema.from_model(rsr)
                        for rsr in researchers]
        }).status(HttpStatus.OK)

    @get("/<int:rsr_id>")
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND,
                              description="Researcher not found",
                              body=ErrorResponseSchema))
    @db.managed_session
    async def get_researcher_by_id(self, req: Request,
                        rsr_id: int,
                        session: AsyncSession) -> Response[ResearcherResponseSchema]:
        """Fetches researcher by id"""
        researcher: Researcher = await Researcher.query(session).filter_by(id=rsr_id).first()
        if researcher is None:
            abort(f"Researcher with id {rsr_id} does not exist.",
                  status_code=HttpStatus.NOT_FOUND)
        return req.res.json({
            "message": "Researcher fetched successfully",
            "status": "success",
            "data": ResearcherOutSchema.from_model(researcher)
        }).status(HttpStatus.OK)

    @get("/<string:rsr_code>")
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND,
                              description="Researcher not found",
                              body=ErrorResponseSchema))
    @db.managed_session
    async def get_researcher_by_code(self, req: Request,
                            rsr_code: str,
                            session: AsyncSession) -> Response[ResearcherResponseSchema]:
        """Fetches researcher by rsr code"""
        researcher: Researcher = await Researcher.query(session).filter_by(code=rsr_code).first()
        if researcher is None:
            abort(f"Researcher with code {rsr_code} does not exist.",
                  status_code=HttpStatus.NOT_FOUND)
        return req.res.json({
            "message": "Researcher fetched successfully",
            "status": "success",
            "data": ResearcherOutSchema.from_model(researcher)
        }).status(HttpStatus.OK)



"""Main exception handler"""
from pyjolt import MediaType, Request, Response, HttpStatus
from pyjolt.controller import produces
from pyjolt.exceptions import (ExceptionHandler,
                               handles,
                               AuthenticationException,
                               AborterException)
from pyjolt.admin import AdminEnterError
from pydantic import ValidationError

class CustomExceptionHandler(ExceptionHandler):
    """Exception handler implementation"""

    @handles(AuthenticationException)
    async def unauthorized(self, req: Request, _: AuthenticationException) -> Response:
        """Handles authentication exceptions"""
        return req.res.json({
            "message": "Login required",
            "status": "error"
        }).status(HttpStatus.FORBIDDEN)

    @handles(AborterException)
    async def aborter_exception(self, req: Request, exc: AborterException) -> Response:
        """Handles aborter exceptions"""
        return req.res.json({
            "message": exc.message,
            "status": exc.status
        }).status(exc.status_code)

    @handles(ValidationError)
    async def validation_error(self, req: Request, exc: ValidationError) -> Response:
        """Handles validation errors"""
        details = {}
        if hasattr(exc, "errors"):
            for error in exc.errors():
                details[error["loc"][0]] = error["msg"]

        return req.response.json({
            "message": "Validation failed.",
            "details": details
        }).status(HttpStatus.UNPROCESSABLE_ENTITY)

    @handles(AdminEnterError)
    @produces(MediaType.TEXT_HTML)
    async def admin_enter_error(self, req: Request, exc: AdminEnterError) -> Response:
        """Handles admin dashboard enter error"""
        print("IN ERROR HANDLER: ", exc.message)
        return (await req.res.html_from_string(
            "<h1>You don't have permission to enter the dashboard</h1>"
        )).status(HttpStatus.OK)



"""
Application user model
"""
from typing import List, TYPE_CHECKING
from sqlalchemy.orm import mapped_column, Mapped, relationship
from pyjolt.admin import register_model
from .base_model import DatabaseModel

if TYPE_CHECKING:
    from .post import Post

@register_model
class User(DatabaseModel):
    """
    User object
    """
    __tablename__ = "users"

    class Meta:
        exclude_from_update_form = ["password", "password_confirm", "created_at"]
        exclude_from_create_form = ["created_at"]
        exclude_from_table = ["password"]
        custom_labels = {
            "fullname": "Full Name",
            "email": "Email Address",
            "password": "Password",
            "password_confirm": "Confirm Password",
        }

    fullname: Mapped[str] = mapped_column(nullable=False)
    email: Mapped[str] = mapped_column(unique=True, nullable=False)
    password: Mapped[str] = mapped_column(nullable=False)

    posts: Mapped[List["Post"]] = relationship(
        back_populates="author",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )



"""
Transition request data schema for state machine extension
"""
from enum import IntEnum, StrEnum, Enum
from typing import Any, Optional
from pydantic import BaseModel, Field

class StateMachineTransitionRequestData(BaseModel):
    """Schema"""
    step: Enum|IntEnum|StrEnum = Field(description="Step to take")
    record_id: int = Field(description="Id of the record to transition")
    data: Optional[Any] = Field(None, description="Any additional data required for the transition")



"""
authentication.py
Authentication module of PyJolt
"""
from abc import ABC, abstractmethod
from typing import (Callable, Optional, Dict,
                    Any, TYPE_CHECKING, Type, cast,
                    TypedDict, NotRequired)
import base64
from datetime import datetime, timedelta, timezone

import bcrypt
import jwt
import binascii
from cryptography.hazmat.primitives.hmac import HMAC
from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import InvalidSignature
from pydantic import BaseModel, Field

from ..exceptions import AuthenticationException, UnauthorizedException
from ..utilities import run_sync_or_async
from ..request import Request
from ..middleware import AppCallableType, MiddlewareBase
if TYPE_CHECKING:
    from ..pyjolt import PyJolt
    from ..response import Response
from ..controller import Controller

REQUEST_ARGS_ERROR_MSG: str = ("Injected argument 'req' of route handler is not an instance "
                    "of the Request class. If you used additional decorators "
                    "make sure the order of arguments was not changed. "
                    "The Request argument must always come first.")
    
USER_LOADER_ERROR_MSG: str = ("Undefined user loader method. Please define a user loader "
                                "method with the @user_loader decorator before using "
                                "the login_required decorator")

class _AuthenticationConfigs(BaseModel):
    """
    Authentication configuration model
    """
    AUTHENTICATION_ERROR_MSG: Optional[str] = Field(
        default="Login required",
        description="Default authentication error message"
    )
    AUTHORIZATION_ERROR_MSG: Optional[str] = Field(
        default="Missing user role(s)",
        description="Default authorization error message"
    )

class AuthConfig(TypedDict):
    """Authentication configurations"""
    AUTHENTICATION_ERROR_MSG: NotRequired[str]
    AUTHORIZATION_ERROR_MSG: NotRequired[str]

class AuthUtils:
    """
    Utility class with useful static methods for authentication
    1. create_signed_cookie_value
    2. decode_signed_cookie
    3. create_password_hash
    4. check_password_hash
    5. create_jwt_token
    6. validate_jwt_token
    """

    @staticmethod
    def create_signed_cookie_value(value: str|int, secret_key: str) -> str:
        """
        Creates a signed cookie value using HMAC and a secret key.

        value: The string value to be signed
        secret_key: The application's secret key for signing

        Returns a base64-encoded signed value.
        """
        if isinstance(value, int):
            value = f"{value}"

        hmac_instance = HMAC(secret_key.encode("utf-8"), hashes.SHA256())
        hmac_instance.update(value.encode("utf-8"))
        signature = hmac_instance.finalize()
        signed_value = f"{value}|{base64.urlsafe_b64encode(signature).decode('utf-8')}"
        return signed_value

    @staticmethod
    def decode_signed_cookie(cookie_value: str, secret_key: str) -> str:
        """
        Decodes and verifies a signed cookie value.

        cookie_value: The signed cookie value to be verified and decoded
        secret_key: The application's secret key for verification

        Returns the original string value if the signature is valid.
        Raises a ValueError if the signature is invalid.
        """
        try:
            value, signature = cookie_value.rsplit("|", 1)
            signature_bytes = base64.urlsafe_b64decode(signature)
            hmac_instance = HMAC(secret_key.encode("utf-8"), hashes.SHA256())
            hmac_instance.update(value.encode("utf-8"))
            hmac_instance.verify(signature_bytes)  # Throws an exception if invalid
            return value
        except (ValueError, IndexError, binascii.Error, InvalidSignature):
            # pylint: disable-next=W0707
            raise ValueError("Invalid signed cookie format or signature.")

    @staticmethod
    def create_password_hash(password: str) -> str:
        """
        Creates a secure hash for a given password.

        password: The plain text password to be hashed
        Returns the hashed password as a string.
        """
        hashed = bcrypt.hashpw(password.encode("utf-8"), bcrypt.gensalt())
        return hashed.decode("utf-8")

    @staticmethod
    def check_password_hash(password: str, hashed_password: str) -> bool:
        """
        Verifies a given password against a hashed password.

        password: The plain text password provided by the user
        hashed_password: The stored hashed password
        Returns True if the password matches, False otherwise.
        """
        return bcrypt.checkpw(password.encode("utf-8"), hashed_password.encode("utf-8"))
    
    @staticmethod
    def create_jwt_token(payload: Dict, secret_key: str, expires_in: int = 3600) -> str:
        """
        Creates a JWT token.

        :param payload: A dictionary containing the payload data.
        :param expires_in: Token expiry time in seconds (default: 3600 seconds = 1 hour).
        :return: Encoded JWT token as a string.
        """
        if not isinstance(payload, dict):
            raise ValueError("Payload must be a dictionary.")

        # Add expiry to the payload
        payload = payload.copy()
        payload["exp"] = datetime.now(timezone.utc) + timedelta(seconds=expires_in)

        # Create the token using the app's SECRET_KEY
        token = jwt.encode(payload, secret_key, algorithm="HS256")
        return token

    @staticmethod
    def validate_jwt_token(token: str, secret_key: str) -> Dict|None:
        """
        Validates a JWT token.

        :param token: The JWT token to validate.
        :return: Decoded payload if the token is valid.
        :raises: InvalidJWTError if the token is expired.
                 InvalidJWTError for other validation issues.
        """
        try:
            # Decode the token using the app's SECRET_KEY
            payload = jwt.decode(token, secret_key, algorithms=["HS256"])
            return payload
        except (jwt.ExpiredSignatureError, jwt.InvalidTokenError):
            raise

class Authentication(MiddlewareBase, ABC):
    """
    Authentication middleware for PyJolt application
    User must implement user_loader method and optionally role_check method
    to define how users are loaded and how roles are checked.
    1. user_loader: should return a user object (or None) loaded from the cookie/jwt/header token
    2. role_check: should check if user has required role(s) and return a boolean
    True -> user has role(s)
    False -> user doesn't have role(s)
    3. Decorators:
        - login_required: to mark route handlers/controllers that require authentication
        - role_required: to mark route handlers/controllers that require specific roles
    """
    configs_name: str = cast(str, None)

    def __init__(self, app: "PyJolt", next_app: AppCallableType) -> None:
        """
        Initilizer for authentication module
        """
        super().__init__(app, next_app)  # type: ignore
        self._configs: dict[str, Any] = {}
        self.authentication_error: str
        self.authorization_error: str

        self._configs = app.get_conf(self.configs_name, {})
        self._configs = self.validate_configs(self._configs, _AuthenticationConfigs)

        self.authentication_error = self._configs["AUTHENTICATION_ERROR_MSG"]
        self.authorization_error = self._configs["AUTHORIZATION_ERROR_MSG"]
        #self._app.add_extension(self) #is this neccessary? - Probably not
    
    async def middleware(self, req: "Request") -> "Response":
        """
        Middleware for authentication
        """
        handler_method: Callable = req.route_handler
        handler_authentication_attributes: Optional[dict[str, Any]] = getattr(handler_method, "_authentication", None)
        controller_authentication_attributes: Optional[dict[str, Any]] = getattr(handler_method.__self__, "_authentication", None) # type: ignore
        if handler_authentication_attributes is None and controller_authentication_attributes is None:
            return await self.next(req)
        if controller_authentication_attributes is not None or handler_authentication_attributes is not None: # type: ignore
            user: Optional[Any] = await run_sync_or_async(self.user_loader, req)
            if user is None:
                #not Authenticated
                raise AuthenticationException(self.authentication_error)
            req.set_user(user)
        controller_roles: list[Any] = controller_authentication_attributes.get("roles", []) if controller_authentication_attributes else []
        handler_roles: list[Any] = handler_authentication_attributes.get("roles", []) if handler_authentication_attributes else []
        roles: list[Any] = list(set(controller_roles + handler_roles))
        if len(roles) == 0: #roles not are specified
            #user is authenticated
            return await self.next(req)
        authorized: bool = await run_sync_or_async(self.role_check, req.user, list(roles))
        if not authorized:
            #not authorized
            raise UnauthorizedException(self.authorization_error, list(roles))
        #user is authenticated and authorized - calls next middleware in chain
        return await self.next(req)

    @abstractmethod
    async def user_loader(self, req: "Request") -> Any:
        """
        Should return a user object (or None) loaded from the cookie
        or some other way provided by the request object
        """

    @abstractmethod
    async def role_check(self, user: Any, roles: list[Any]) -> bool:
        """
        Should check if user has required role(s) and return a boolean
        True -> user has role(s)
        False -> user doesn't have role(s)
        """

def login_required(handler: "Callable|Type[Controller]") -> "Callable|Type[Controller]":
    """
    Decorator for login required
    """
    setattr(handler, "_authentication", {
        "required": True
    })
    return handler

def role_required(*roles) -> Callable[[Callable|Type[Controller]], Callable|Type[Controller]]:
    """
    Decorator for role required
    """
    def decorator(handler: "Callable|Type[Controller]") -> "Callable|Type[Controller]":
        attributes: dict[str, Any] = getattr(handler, "_authentication", {})
        attributes["roles"] = list(roles)
        attributes["required"] = True
        setattr(handler, "_authentication", attributes)
        return handler
    return decorator



"""Admin dashboard extension"""
from typing import Type
from pyjolt import Request
from pyjolt.admin import AdminDashboard
from pyjolt.caching.cache import Cache
from pyjolt.email import EmailClient
from pyjolt.database.sql.declarative_base import DeclarativeBaseModel

class AdminExtension(AdminDashboard):
    """Admin dashboard extension"""

    async def has_enter_permission(self, req: Request) -> bool:
        return True

    async def has_create_permission(self, req: Request, model: Type[DeclarativeBaseModel]) -> bool:
        return True

    async def has_delete_permission(self, req: Request, model: Type[DeclarativeBaseModel]) -> bool:
        return True

    async def has_update_permission(self, req: Request, model: Type[DeclarativeBaseModel]) -> bool:
        return True

    async def has_view_permission(self, req: Request, model: Type[DeclarativeBaseModel]) -> bool:
        return True

    async def has_cache_permission(self, req: Request, cache: Cache) -> bool:
        return True

    async def email_recipient_query(self, req: Request, query: str, client: EmailClient) -> list[tuple[str, str]]:
        """
        Email recipient query implementation
        """
        return [
            ("Marko Šterk", "marko.sterk@izum.si"),
            ("Andrej Korošec", "andrej.korosec@izum.si"),
            ("Andrej Zidarič", "andrej.zidaric@izum.si"),
            ("Janja Zorman", "janja.zorman@izum.si"),
            ("Filip Pasarič", "filip.pasaric@izum.si")
        ]
    
    async def has_files_permission(self, req):
        return True

admin_extension: AdminExtension = AdminExtension()



"""Post model for blog posts"""
from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional
from sqlalchemy.orm import mapped_column, Mapped, relationship
from sqlalchemy import ForeignKey, Text, event
from pyjolt import Request
from pyjolt.database.sql import AsyncSession
from pyjolt.admin import register_model
from pyjolt.admin.input_fields import TagsInput
from pyjolt.utilities import to_kebab_case
from .base_model import DatabaseModel
from app.api.schemas.post_schemas import (PostsQuery, PostInSchema,
                                          any_tag_in_csv_condition,
                                          PostOutSchema)

if TYPE_CHECKING:
    from .user import User

@register_model
class Post(DatabaseModel):
    """Post model"""
    __tablename__ = "posts"

    class Meta:
        exclude_from_create_form = ["created_at", "slug"]
        exclude_from_update_form = ["created_at", "slug"]
        exclude_from_table = ["slug", "content_slv"]
        order_table_by = ["created_at DESC"]
        custom_labels = {
            "title_eng": "Title (EN)",
            "title_slv": "Title (SLV)",
            "content_eng": "Content (EN)",
            "content_slv": "Content (SLV)",
            "id": "ID"
        }
        form_fields_order = ["title_slv", "title_eng", "content_slv", "content_eng", "active", "tags_list"]
        custom_form_fields = [
            TagsInput(id="tags_list", name="tags_list", label="Tags", as_string=True)
        ]
        create_validation_shema = PostInSchema
        update_validation_shema = PostInSchema

    title_eng: Mapped[str] = mapped_column()
    title_slv: Mapped[str] = mapped_column()
    slug: Mapped[str] = mapped_column(unique=True)
    content_eng: Mapped[str] = mapped_column(Text, nullable=False)
    content_slv: Mapped[str] = mapped_column(Text, nullable=False)
    active: Mapped[bool] = mapped_column(default=True)
    tags_list: Mapped[str] = mapped_column(nullable=False, default="")

    author_id: Mapped[int] = mapped_column(
        ForeignKey("users.id", ondelete="CASCADE"),
        nullable=False,
    )

    author: Mapped[User] = relationship(back_populates="posts")

    @property
    def tags(self) -> list[str]:
        """Tags of the post"""
        return [t for t in (self.tags_list.split(",") if self.tags_list else []) if t]

    @tags.setter
    def tags(self, value: list[str]):
        norm = [t.strip().lower() for t in value if t and t.strip()]
        self.tags_list = ",".join(norm)

    @classmethod
    async def query_posts(cls, session: AsyncSession,
                          query_data: PostsQuery) -> dict[str, Any]:
        """Performs query for posts"""
        conds = []
        if query_data.active is not None:
            conds.append(cls.active == query_data.active)
        if query_data.created_at is not None:
            conds.append(cls.created_at >= query_data.created_at)
        if query_data.tags:
            conds.append(any_tag_in_csv_condition(query_data.tags, Post))
        if query_data.user_id is not None:
            conds.append(cls.author_id == query_data.user_id)

        results: dict[str, Any] = await Post.query(session).filter(
            *conds
        ).paginate(page=query_data.page, per_page=query_data.per_page)
        results["items"] = [PostOutSchema.from_model(post)
                          for post in results["items"]]
        return results
    
    async def admin_create(self, req: "Request", new_data: dict[str, Any],
                           session: "AsyncSession"):
        """Saves the post from admin interface. Sets author_id from request user."""
        new_data["tags_list"] = ",".join(new_data["tags_list"])
        for key, value in new_data.items():
            setattr(self, key, value)
        self.author_id = req.user.id
        session.add(self)
    
    #this hook or "before_update" can be used for modifying the tags_list from list[str] to str
    async def admin_update(self, req: "Request", new_data: dict[str, Any],
                           session: "AsyncSession"):
        tags: Optional[str] = None
        if new_data.get("tags_list"):
            tags = ",".join(new_data["tags_list"])
            new_data["tags_list"] = tags
        for key, value in new_data.items():
            setattr(self, key, value)
        session.add(self)

@event.listens_for(Post, "before_insert")
def set_slug_before_insert(mapper, connection, target: Post):
    if target.title_eng:
        target.slug = to_kebab_case(target.title_eng)


@event.listens_for(Post, "before_update")
def set_slug_before_update(mapper, connection, target: Post):
    # Only update if title_eng changed or slug is empty
    if target.title_eng and (not target.slug or target.slug != to_kebab_case(target.title_eng)):
        target.slug = to_kebab_case(target.title_eng)
    # if target.tags_list:
    #     target.tags_list = ",".join(target.tags_list)



"""Projects API"""
from typing import Any
from pyjolt import Request, Response, HttpStatus, MediaType, abort
from pyjolt.controller import (Controller, path,
                               get, produces, open_api_docs,
                               Descriptor)
from pyjolt.database.sql import AsyncSession

from app.extensions import db
from app.api.models.project import Project
from app.api.schemas.base_schema import ErrorResponseSchema
from app.api.schemas.project_schemas import (ProjectOutSchema,
                                             ProjectQuerySchema,
                                             ProjectsQueryResponseSchema,
                                             ProjectResponseSchema)

@path("/api/v1/projects")
class ProjectsApi(Controller):
    """Projects api controller"""

    @get("/")
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    async def get_projects(self, req: Request,
                    session: AsyncSession) -> Response[ProjectsQueryResponseSchema]:
        """Fetches all projects"""
        query_data: ProjectQuerySchema = ProjectQuerySchema.model_validate(req.query_params)
        result: dict[str, Any] = await Project.query_projects(session, query_data)

        return req.res.json({
            "message": "Query successful",
            "status": "success",
            "data": result
        }).status(HttpStatus.OK)

    @get("/<int:project_id>")
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND,
                              description="Project not found",
                              body=ErrorResponseSchema))
    @db.managed_session
    async def get_project_by_id(self, req: Request,
                          project_id: int,
                          session: AsyncSession) -> Response[ProjectResponseSchema]:
        """Fetches project by id"""
        project: Project = await Project.query(session).filter_by(id=project_id).first()
        if project is None:
            abort(f"Project with id {project_id} not found.",
                  status_code=HttpStatus.NOT_FOUND)

        return req.res.json({
            "message": "Project fetched successfully",
            "status": "success",
            "data": ProjectOutSchema.from_model(project).model_dump()
        }).status(HttpStatus.OK)

    @get("/<string:project_code>")
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND,
                              description="Project not found",
                              body=ErrorResponseSchema))
    @db.managed_session
    async def get_project_by_code(self, req: Request,
                                  project_code: str,
                                  session: AsyncSession) -> Response[ProjectResponseSchema]:
        """Fetches project by project code"""
        project: Project = await Project.query(session).filter_by(code=project_code).first()
        if project is None:
            abort(f"Project with code {project_code} does not exist",
                  status_code=HttpStatus.NOT_FOUND)

        return req.res.json({
            "message": "Project fetched successfully",
            "status": "success",
            "data": ProjectOutSchema.from_model(project).model_dump()
        }).status(HttpStatus.OK)



"""Controller subpackage"""

from .controller import Controller, Descriptor, path
from .decorators import (get, post, delete, patch, put,
                         before_request, after_request,
                         produces, consumes, open_api_docs,
                         cors, no_cors, socket, development)

__all__ = ["Controller", "path", "get", "post", "put",
           "patch", "delete", "consumes",
           "produces", "Descriptor", "open_api_docs",
           "before_request", "after_request", "cors", "no_cors",
           "socket", "development"]



"""CLI utility script"""


#Available cli command for db migrations
#Database models must be imported into the create_app factory function in order
#for the migration module to detect the correctly
#uv run cli.py db-init
#uv run cli.py db-migrate
#uv run cli.py db-upgrade

if __name__ == "__main__":
    ##CLI interface for application
    from app import Application
    app = Application()
    app.run_cli()



"""
Import all extensions and initilize
After that import them into the create_app method and initilize with application
"""
from pyjolt.database.sql import SqlDatabase
from pyjolt.database.migrate import Migrate

db: SqlDatabase = SqlDatabase()
migrate: Migrate = Migrate(db)

__all__ = ["db", "migrate"]
